%%% SYSTEM CORE

% This file should be loaded first

% Operators:
% -x suffix, hops right
% #x suffix, doesn't hop
% x~ prefix, doesn't hop
% x@ prefix, hops left

% - op(200,fy,'-') default, hops right
:- op(200,fy,'#').		% does not hop
:- op(200,yf,'~').		% does not hop
:- op(200,yf,'=').		% hops left
:- op(950,xfx,if).

%%% Machine Architecture
%%%
%%% (c) Sandiway Fong, University of Arizona 2013, 2014, 2015

:- multifile parameterization/2.
:- multifile parameter_values/2.
:- discontiguous parameterization/2.
:- discontiguous parameter_values/2.

:- dynamic print_flag/0.
:- dynamic pending_step/0.
:- dynamic streams/2.		% defined in prolog example files only
:- dynamic stream/2.		% defined in prolog example files only

%% RUN
%%   run(List/Name,SO,S) 	list of heads or predefined stream name, 
%%				produces syntactic object (SO) and spellout (S)
%%   run(List/Name)
%%   run(Streams) 		predefined list of stream names (Streams)
%%   run_check(Streams)		run all examples defined by name Streams
%% PARAMETERIZATION
%%   substream_stack TF substream stack transfer to mainstream
%%   pmerge_flag	alternate order spellout option controlled by flag
%%			true/false
%%   le(on/off/pc)	turns on/off local extent (LE) (stack:b) marking on the
%%			stack
%%			3rd option pc: phase completeness (PC) determines LE
%%			marking
%%   inactive_push TF      DC merge, nonhead gets pushed with inactive wrap onto stack
%%   search(so/stk)	search SO for unvalued uF
%% 			or search stack for inactive wrapped unvalued uF
%%   drop_stack TF	drop stack elements 2 boundaries down
%%   stack_access	operations like probe-goal access the stack
%%   (local/global)	controls whether stack beyond the local extent (LE)
%%			can be accessed
%%   subphrase_check 
%%   (off,subsume,	subsume:  subconstituent is subsumed
%%    on_fail,		on_fail: fail if subconstituent is found
%%    subsume_ss, 	subsume_ss: only subconstituent from a substream is subsumed
%%    on_fail_ss, 	on_fail_ss: fail if subconstituent from a substream is found
%%    off)		off: no check
%%   stack_subc		opush substitutes subconstituent on stack with fresh one
%%			marking through old one
%%   c_stack_search	c searches the stack
%%   (normal,inactive)     normal: respect inactive wrapper
%%			inactive: stack search can see past inactive wrapper
%%
%%   vprint TF		controls stack printing
%%			(vertical/horizontally on one line)
%%   iprint TF		controls inactive(_) wrapper stack printing
%%			(on/off)
%%
%%   ?- debug(probegoal). 	to print probe/goal information
%% CHANGE NOTES
%%   uses leading _ to denote unpronounced copies
%%   has a spell-out implementation
%%   implements v~ (Sobin, 2012) for TH/EX and there-insertion
%%   pop X at YP if X has no unvalued uF and no interpretable wh-feature
%%   keeps X with interpretable-wh on stack until CP
%%   implements Pesetsky's & Torrego 2001 properly
%%   T with uv
%%   DC Binding
%%   pair-merge (alternate order controlled by pmerge_flag)
%%   implements Phase Completeness DC Binding

%% machine parameterizations

%% substream_stack transfer flag
:- dynamic substream_stack/0.
parameterization(substream_stack,
		 'Allow transfer of substream stack back to mainstream').
substream_stack.
% substream_stack :- fail.	% no transfer of stack contents

%% stack update parameter
parameterization(stack_update_larger,
		 'Replace updated subconstituent of stack element').
:- dynamic stack_update_larger/0.
% stack_update_larger.

%%  pair-merge(X,Y) flag
%%  on: allow spellout as XY or YX
%%  off: spellout as XY only
:- dynamic pmerge_flag/0.
parameterization(pmerge_flag,'Allow spellout of pair-merge in either order').
pmerge_flag :- fail.
% pmerge_flag.

%% inactive_push flag
%% on: push nonhead of DC pair onto the stack, wrapped in inactive(_)
%% off: no push
:- dynamic inactive_push/0.
parameterization(inactive_push,'Push nonhead of DC merge onto the stack inactive(_)').
inactive_push.
% inactive_push :- fail.

%% search(so/stk) flag
%% so: search SO
%% stk: search stack
:- dynamic search/1.
parameterization(search(_),'Search SO or stack for unvalued uF to be released').
parameter_values(search(_),[so,stk,off]).
search(stk).
% search(so).
% search(off).

%% drop_stack flag
%% on: drop stack items that are two phase boundaries down
%% off: no drop
:- dynamic drop_stack/0.
parameterization(drop_stack,
		 'Drop stack items that are two phase boundaries down').
drop_stack :- fail.
% drop_stack.

%% stack_access flag
%% local: not past a boundary (b) mark
%% global: ignore boundary marks
:- dynamic stack_access/1.
parameterization(stack_access(_),
		 'Local or global access to elements of the stack').
parameter_values(stack_access(_),[local,global]).
stack_access(local).
% stack_access(global).

%% Define: phase completeness (PC) =
%% 	  max projected phase + no unvalued uF at head or edge of phase
%% Three options: pc, on, off
:- dynamic le/1.
parameterization(le(_),
		 'Local Extent boundary conditions, used to trigger stacking of unvalued uF SOs').
parameter_values(le(_),[on,pc,off]).
le(on). % phase boundary = max projected phase
% le(pc).				% phase boundary = phase completeness
% le(off). % no phase boundary checking

%% subphrase_check parameter: on push to stack, this controls checking
%   subsume:  subconstituent is subsumed
%   on_fail: fail if subconstituent is found
%   subsume_ss: only a subconstituent from a substream is subsumed
%   on_fail_ss: fail if subconstituent from a substream is found
%   off: no check
:- dynamic subphrase_check/1.
parameterization(subphrase_check(_),
	'Push element onto stack, first check if subconstituent is already there').
parameter_values(subphrase_check(_),
		 [subsume,on_fail,subsume_ss,on_fail_ss,off]).
subphrase_check(subsume_ss).

%%   c_stack_search	c searches the stack
%%   (normal,inactive)     normal: respect inactive wrapper
%%			inactive: stack search can see past inactive wrapper
%%
%% In: Who1 thinks that c_eQ Mary likes him1, c_eQ needs to see inactive(who)
:- dynamic c_stack_search/1.
parameterization(c_stack_search(_),
		 'c stack search normal or can see into inactive wrapped elements').
parameter_values(c_stack_search(_),[normal,inactive]).
c_stack_search(inactive).

:- dynamic occ_new_flag/0.
parameterization(occ_new_flag,'if already old, create new, otherwise keep old').
% occ_new_flag.
% occ_new_flag :- fail.

:- dynamic subc/0.
parameterization(stack_subc,'if existing subconstituent found on stack, mark old and substitute new on stack').
% stack_subc.
stack_subc :- fail.

%% vertical stack printing flag
%% on: print stack as a vertical list
%% off: print stack as a horizontal list
:- dynamic vprint/1.
parameterization(vprint(_),'Print stack as vertical list').
parameter_values(vprint(_),[on,off]).
vprint(on).			
% vprint(off).			

%% inactive wrapper printing
%% on: prints inactive(_) for stack.
%% off: omits inactive(_)
:- dynamic iprint/1.
parameterization(iprint(_),'Print inactive(_) wrappers').
parameter_values(iprint(_),[on,off]).
% iprint(on).			
iprint(off).			

% supply stream (list of heads) or use streams named below
run(Name,SO,S) :-
    nb_setval(fail_chain,0),
    resetC,
    bumpD,
    init_stats,
    retractall(print_flag), 
    retractall(pending_step), 
    resetStep,
    (stream(Name,L) -> true ; L = Name),
    mkStream(L,Str),
    printState(Str,[],[],0),
    machine(Str,[],[],[],Stk,SO),
    wff(Stk),		% all uF must have been valued
    ci(SO),			% conceptual intentional (CI) interface
    output(''),
    print_spellout(SO,S),
    output(''),
    assert(pending_step).
	
run(Name) :-
%	attach_console, spy(wff),
%	attach_console, spy(draw_tree),
    start_encoder,
    print_parameters,
    (streams(Name,L) ->
	 member(S,L), 
	 format(string(Out),'Example: ~w',[S]), 
	 output(Out),
	 run(S,_,_),
	 fail 
     ;  run(Name,_,_),
	fail).
run(_) :-
    parseC(N),
    (N == 0 ->
	 output_strong("No derivations found.")
     ; (N == 1 ->
	    output_strong("One derivation found.")
	;  format(string(Out),'~w derivations found.',[N]),
	   output_strong(Out))),
    output("Computational Cost:"),
    print_stats,
    output_link,
    output(''),
    finish_encoder.

%% run_nc(Name) variant of run(Name)
%% NC = no clearing of console,
%% no parameter or computational cost display
run_nc(Name) :-
    (streams(Name,L) ->
	 member(S,L), 
	 format(string(Out),'Example: ~w',[S]), 
	 output(Out),
	 run(S,_,_),
	 fail 
     ;  run(Name,_,_),
	fail).
run_nc(_) :-
    parseC(N),
    (N == 0 ->
	 output_strong("No derivations found.")
     ; (N == 1 ->
	    output_strong("One derivation found.")
	;  format(string(Out),'~w derivations found.',[N]),
	   output_strong(Out))),
    output_link,
    output('').

% for verifying examples
% Name = group name for examples, e.g. basic

run_check(Name) :- retractall(silent), asserta(silent), run_check1(Name), fail.
run_check(_) :-    retractall(silent).

run_check1(Name) :-
    (streams(Name,SL) ->
	 true
     ;   format(string(Out),'Section: ~w ERROR: does not exist!',[Name]),
	 force_output(Out),
	 fail),
    member(S,SL),
    (stream(S,_) ->
	 true
     ;   format(string(Out),'Example: ~w ERROR: not found!',[S]),
	 force_output(Out),
	 fail),
    (run(S,_,Spellout) *->
	(spelled(S,Spellout) ->
	     format(string(Out),'Example: ~w ~w OK',[S,Spellout])
	 ;   (spelled(S,Spelled) ->
		  format(string(Out),'Example: ~w:~w ERROR: ~w',[S,Spelled,Spellout])
	      ;   format(string(Out),'Example: *~w ERROR: parsed as ~w',[S,Spellout])))
     ; (spelled(S,Spelled) ->
	    format(string(Out),'Example: ~w:~w ERROR: no parse',[S,Spelled])
	;   format(string(Out),'Example: ~w:no parse OK',[S]))),
    force_output(Out).

% development

% c-T uF phi, tns
% T uF phi[person], no tns
% edge allows it to enter derivations
% In {H, α}, where H is a Lexical Item, H is the label.
% If α is a term of β, the label of {α, β} is the label of β. (Internal Merge)

if(X,Y) :- Y -> X ; true.

%% Spell-out section

prettify_list([],[]).
prettify_list([X|L1],[Y|L2]) :- prettify(X,Y), prettify_list(L1,L2).

prettify(X,Y) :-
    (atomic(X) ->
	 Y = X
     ; functor(X,_,N),
       arg(N,X,Nth),
       (var(Nth) ->
	    X =.. L,
	    append(NewL,[_],L),
	    Y =.. NewL
	;   Y = X)).

print_list([]).
print_list([X|L]) :-
    prettify(X,Xp),
    format('~w ',[Xp]),
    print_list(L).

% old
% print_list([X|L]) :- format('~w ',[X]), print_list(L).

print_list2(L) :- print_list2(L,[]).

print_list2([],_). 
print_list2([X|L],Vs) :- 
    (X = t(T,P,Occ) ->
	 Y = t(T,P),
	 (varIn(Occ,Vs) ->
	      format('<del>~w</del> ',[Y]), Vs2 = Vs
	  ;   format('~w ',[Y]), Vs2 = [Occ|Vs])
     ; (X = v(C,Occ) ->
	    (varIn(Occ,Vs) ->
		 format('<del>~w</del> ',[C]), Vs2 = Vs
	     ;  format('~w ',[C]), Vs2 = [Occ|Vs])
	;  format('~w ',[X]), Vs2 = Vs)),
    print_list2(L,Vs2).

varIn(X,[Y|_]) :- X == Y, !. 	% red
varIn(X,[_|L]) :- varIn(X,L).

% affix_hop(L,Lp,L_end), L_end seems appended to the end of the list
affix_hop(L1,L2,L3) :-
    affix_hopR(L1,L1p,L3),
    affix_hopL(L1p,L2).

affix_hopR([],L,L).
affix_hopR([A|L],L1,L5) :- 
    (suffix_hopR(A) ->
	 (L == [] ->		% affix with nothing to hop over
	      L1 = [A]
	  ;  affixes_hopR([A|L],[],As,L2),
	     find_head(L2,Prefixes,H,L3), % hop over H
	     append(Prefixes,[H|As],PHS),
	     append(PHS,L4,L1),
	     affix_hopR(L3,L4,L5))
     ;  L1 = [A|L2],
	affix_hopR(L,L2,L5)).

affix_hopL(L1,L2) :-
    reverse(L1,L1r),
    affix_hopL2(L1r,L2r),
    reverse(L2r,L2).

% reverse order, prefix_hopL 
affix_hopL2([],[]).
affix_hopL2([A|L],L1) :-
    (prefix_hopL(A) ->
	 (L == [] ->
	      L1 = [A]
	  ; affixes_hopL([A|L],[],As,L2),
	    find_head(L2,Prefixes,H,L3),
	    append(Prefixes,[H|As],PHS),
	    append(PHS,L4,L1),
	    affix_hopL2(L3,L4))
     ; L1 = [A|L2],
       affix_hopL2(L,L2)).

% Lp = input after sequence of affixes
% Bs = sequence of affixes
affixes_hopR([X|L],As,Bs,Lp) :-
    (suffix_hopR(X) ->
	 affixes_hopR(L,[X|As],Bs,Lp)
     ;  (prefix(X) ->
	     Lp = [X|L2],
	     affixes_hopR(L,As,Bs,L2)
	 ;  Bs = As,
	    Lp = [X|L])).

affixes_hopL([X|L],As,Bs,Lp) :-
    (prefix_hopL(X) ->
	 affixes_hopL(L,[X|As],Bs,Lp)
     ;  (prefix(X) ->
	     Lp = [X|L2],
	     affixes_hopL(L,As,Bs,L2)
	 ;  Bs = As,
	    Lp = [X|L])).

suffix(X) :- suffix_nohop(X) ; suffix_hopR(X).
prefix(X) :- prefix_nohop(X) ; prefix_hopL(X).
suffix_nohop(#_).
suffix_hopR(-_).
prefix_nohop(_~).
prefix_hopL(_=).

find_head([X|L],[],X,L) :- \+ (suffix(X) ; prefix(X)), !.
find_head([X|L],[X|Ps],H,L2) :-
    find_head(L,Ps,H,L2).
	
% based on portray/1, uses difference lists
% L1 is the list of heads on the frontier of the SO
% pair-merge ordering depends on the basetype of X
frontier(set(List),Occ1,L1,L3) :- % set
	!,			  %red
	frontierL(List,Occ1,L1,L3).
frontier(pm([X,Y]),Occ1,L1,L3) :-	% pair merge
    label_basetype(X,T), 
    (headInitial(T) ->		% <X,Y> when X is head-initial
	 frontier(X,Occ1,L1,L2),
	 frontier(Y,Occ1,L2,L3)
     ;   frontier(Y,Occ1,L1,L2),
	 frontier(X,Occ1,L2,L3)).
frontier(pm([X,Y]),Occ1,L1,L3) :-	% alternate order
    pmerge_flag,			% machine parameter
    label_basetype(X,T), 
    (headInitial(T) ->		% <X,Y> when X is head-initial
	 frontier(Y,Occ1,L1,L2),
	 frontier(X,Occ1,L2,L3)
     ;   frontier(X,Occ1,L1,L2),
	 frontier(Y,Occ1,L2,L3)).
frontier(so(C,UFs,_,Fs,SOs),Occ1,L1,L3) :-
    propagateOcc(Occ1,Occ), 
    frontierC(C,UFs,Fs,SOs,Occ,L1,L2), 
    frontierL(SOs,Occ,L2,L3).

% outputs only heads without occurrence flag set
frontierC(C,UFs,Fs,SOs,Occ,L1,L2) :-
    (member(occ(X),Fs) ->
	 (var(X) ->
	      (SOs == [] ->
		   (portray_emptyhead(C,UFs,Fs,L3) ->
			append(L3,L2,L1)
		    ;  L1 = [C|L2])
	       ;  L2 = L1),
	      X = 1		% occ(1)
	  ;  (occ_exception(C) ->
		  (portray_emptyhead(C,UFs,Fs,L3)->
		       append(L3,L2,L1)
		   ;  L1 = [C|L2])
	      ;  L2 = L1, 
		 Occ = 1))
     ;  (var(Occ) ->
	     (SOs == []	->	% empty head
		  (portray_emptyhead(C,UFs,Fs,L3) ->
		       append(L3,L2,L1)
		   ;  L1 = [C|L2])
	      ; L2 = L1)
	 ; L2 = L1)
    ).

% T/v may need multiple spellout
%%% What examples?
%%% NEW: Tinf raised by c_rel, lower Tinf should not be spelled out
%%% the baker Tinf c_rel PRO Tinf PRO v* place you 's trust in

occ_exception(X) :- X \== 'Tinf', (type(X,'T') -> true ; type(X,v)).

frontierL([],_Occ,L,L).
frontierL([SO|SOs],Occ,L1,L3) :- 
    frontier(SO,Occ,L1,L2), 
    frontierL(SOs,Occ,L2,L3).

% add -m suffix to matrix c_Q
matrixC(so(C,UFs,IFs,Fs,SOs),so(CM,UFs,IFs,Fs,SOsp)) :-
    type(C,c),
    (SOs == [] -> matrix(C,CM), SOsp = SOs ; CM = C, matrixCL(SOs,SOsp)),
    !. 				% red
matrixC(X,X).

matrix(c_Q,c_Qm).

matrixCL([],[]).
matrixCL([SO|SOs],[SOp|SOsp]) :-
    (matrixC(SO,SOp) -> true ; SOp = SO),
     matrixCL(SOs,SOsp).

% Ways to spell out a head H
% 1. disappears when no_spellout(H) is true
% 2. H H2 spellout as H3 (H2 = next head), then replace H H2 by H3
% 3. H spells out as H3 when followed by H2, then it's H3 followed by spellout of H2...
% 4. H has agreement features or morph(type(H)) is true,
%    if realize_head(H,H2), it's H2
%    if realize_head(H,_) is not true, it disappears
% 5. if realize_head(H,H2), it's H2 otherwise 6. applies
% 6. none of the above applies, H spells out as H

spellout_heads([],[]).
spellout_heads([H|T],L) :-
    no_spellout(H),
    !, 				% red
    spellout_heads(T,L).
spellout_heads([X,Y|_],_) :-
	block_heads(X,Y),
	!,			% red
	printFailFormat('Crash: sequence \'~s ~s \' cannot be spelled out',[X,Y]),
	fail.
spellout_heads([X,Y|L],Lp) :-
    realize_heads2(X,Y,L1),	% X followed by Y realized by L1
    append(L1,L2,Lp),
    !,				% red
    spellout_heads(L,L2).
spellout_heads([X,Y|L],Lp) :-
    realize_heads(X,Y,L1),	% X is realized as L1 if followed by Y
    append(L1,L2,Lp),
    !,				% red
    spellout_heads([Y|L],L2).
spellout_heads([H|T],L) :-
    (morpheme_withFs(H) ; type(H,V), morph(V)),
    !,				% red
    (realize_head(H,S) ->
	 (S = [X,Y] -> L = [X,Y|Lp] ; (S == [] -> L = Lp ; L = [S|Lp]))
     ;  L = Lp),
    spellout_heads(T,Lp).
spellout_heads([H|T],L) :-
    realize_head(H,S),		% realize single head H
    !,				% red
    (S = [X,Y] -> L = [X,Y|Lp] ; L = [S|Lp]),
    spellout_heads(T,Lp).
spellout_heads([H|T],[H|L]) :- spellout_heads(T,L).

% stage 2 head spellout
spellout2_heads([],[]).
spellout2_heads([H|T],L) :-
    realize2_head(H,S),
    !,				% red
    (S = [X,Y] -> L = [X,Y|Lp] ; (S == [] -> L = Lp ; L = [S|Lp])),
    spellout2_heads(T,Lp).
spellout2_heads([X,Y|L],Lp) :-
    realize2_heads(X,Y,L1),	% X is realized as L1 if followed by Y
    append(L1,L2,Lp),
    !,				% red
    spellout2_heads([Y|L],L2).
spellout2_heads([H|T],[H|L]) :- spellout2_heads(T,L).

% last stage
spell([],[]).
spell([X],[X]).
spell([X,Y|L],Lp) :- 
    (suffix(Y) -> 
	 arg(1,Y,Y1),
	 (spell_out(Y1,X,Z) ->
	      (Z == nospellout -> printNospell(X,Y), fail ; true)
	  ;  Z = X),		% deletes suffix
	 spell([Z|L],Lp) 
     ;
     (prefix(X) ->
	  arg(1,X,X1),
	  spell([Y|L],[Yp|L2p]),
	  atomic_list_concat([X1,'-',Yp],XYp),
	  Lp = [XYp|L2p]
      ;   Lp = [X|L2], 
	  spell([Y|L],L2))).

%% Machine

:- dynamic lastSSStack/1.

%%% NEW: substream stack merge needs to be delayed until external
%%% merge of computed SSO
% stream x stack x SO -> stream' x stack' x SO'
machine([Substr|Str],Stk,SO,Strp,Stkp,SOp) :-
    Substr = [_|_],
    !,			% green
    step(X),
    output("Begin substream"),
    printState(Substr,[],[],X),
    machine(Substr,[],[],[],SStk,SSO), % keep substream stack
    printSSOInsert(SSO),
    (substream_stack ->
	 (SStk \== [] ->
	      add_substreamFL(SStk,SStk1), % deletes b
%	      append(SStk1,Stk,Stk2),
	  addStackDelay(SStk1,SSO,SSO1),
	  Stk2 = Stk
	 ;
	  Stk2 = Stk,
	  SSO1 = SSO)
    ;
     Stk2 = Stk,
     SSO1 = SSO),
    step(X2),
    printState([SSO1|Str],Stk2,SO,X2),
    add_substreamF(SSO1,SSOp),
    machine([SSOp|Str],Stk2,SO,Strp,Stkp,SOp).
machine(Str,Stk,SO,Str2,Stk2,SO2) :-
    (action(Str,Stk,SO,Strp,Stkp,SOp,Step) *->
	   printState(Strp,Stkp,SOp,Step),
           machine(Strp,Stkp,SOp,Str2,Stk2,SO2)
     ; (Str == [] ->
	    Str2 = Str, Stk2 = Stk, SO2 = SO
	;  printNoAction(Str,SO),
	   fail)).

%%% NEW delay is a feature with a stack
addStackDelay(SStk1,SO,SOp) :-
	addF(delay(SStk1),SO,SOp),
	printSSInsert(SStk1).

addF(F1,so(LB,U,I,F,CS),so(LB,U,I,[F1|F],CS)).

% use delayed stack stored on X (on merge from stream)
mergeDelayedStack(X,ST,STp) :-
	(f(X,delay(SStk)) ->
	 append(SStk,ST,STp),
	 printMergeDelayedStack(SStk)
	; STp = ST).

%%% NEW
%%% your trust in who_rel baker (relative clause)
%%% you + 's => b (LE boundary) from the spec
%%% merged with main stream prevents access to who_rel
add_substreamFL([b],[]) :- !. 	% red
	
add_substreamFL([],[]).
add_substreamFL([X|L],[Y|Lp]) :- add_substreamF(X,Y), add_substreamFL(L,Lp).

% adds feature substream(OL) if not present
% if present, extend open list OL length
% inactive(_) wrapped stack elements are not marked
% NOT USED YET
%   lengthOL(OL,N) indicates the depth of the substreaming

add_substreamF(pm([X,Y]),pm([Xp,Y])) :-
    !,				% green
    add_substreamF(X,Xp).
add_substreamF(X,X) :-
    functor(X,inactive,1),
    !.				% green
%%% NEW - e42b Binding [once(John),b]
add_substreamF(once(X),once(Xp)) :-
    !,				% green
    add_substreamF(X,Xp).
add_substreamF(X,X) :-
    boundaryMark(X),
    !.				% green
%%% END NEW
add_substreamF(so(C,UF,IF,F,L),so(C,UF,IF,Fp,L)) :-
    (member(substream(OL),F) ->
     extendOL(OL),
     Fp = F			% NEW: needed oo/w Fp is a variable
    ; Fp = [substream(_)|F]).

boundaryMark(b).

%% Phi feature support

% phi-complete if no phi-feature xcluded
phi_complete(X) :- uF(X,f(phi,Vs,_)), (var(Vs) -> true ; \+ (member(V,Vs), V == x)).

% phi-feature valuation
% default [3,sg,n] if W = [_,_,_]

value_phi(V,W) :-
    (var(V) ->
	 ((W = [P,N,G],
	   (var(P) -> P=3 ; true),
	   (var(N) -> N=sg ; true),
	   (var(G) -> G=n ; true)) ->
	      V = W
	  ;  V = W)
     ; valueL(V,W)).

valueL([],[]).
valueL([F1|Fs1],[F2|Fs2]) :-
    value1(F1,F2),
    !,				% red
    valueL(Fs1,Fs2).

% 1st argument
% V  variable op:unify
% x  skip
% NV  unified(X) 
% NV nonvar   unify
value1(V,X) :- var(V), !, V = X. % red
value1(x,_) :- !.		 % red
value1(unified(X),X).
value1(X,X).

unify_phi(V,W) :- var(W), !, W = V.
unify_phi(V,W) :- unifyL(V,W,Flag), Flag == unified.

% X is from a probe, Y a goal
% X Y  Action
% _ x  skip
%   V  link variable V to X assumed to be a variable,
%      once linked, V is no longer considered a variable
% x _  fail?
%   NV NV = X 
unifyL([],[],_).
unifyL([X|L1],[Y|L2],Flag) :-
    unify1(X,Y,Fg), updateFlag(Flag,Fg), unifyL(L1,L2,Flag).

updateFlag(F,Fg) :- ((var(F), Fg == unified) -> F = unified ; true).

% X none/unified flag
% V             W
% -		x (no value) 	flag: none
% x		variable	flag: none
% -		unified(v)	flag: unified
% v		v		flag: unified
unify1(V,W,X) :-
	(W == x ->
	     X = none
	 ;  (var(W) ->
		 (V == x ->
		      X = none
		  ;  W = unified(V),
		     X = unified)
	     ;  W = V,
		X = unified)).

% list of feature specifications, e.g.
% iF(theta)
unifyFL([],_,_).
unifyFL([FS|L],X,Y) :-
    selectF(FS,X,F),
    selectF(FS,Y,F),
    unifyFL(L,X,Y).

selectF(iF(FN),X,F) :-
    iF(X,F),
    functor(F,FN,_),
    !. 				% green
selectF(uF(FN),X,F) :-
    uF(X,F),
    functor(F,FN,_),
    !. 				% green
selectF(f(FN),X,F) :-
    f(X,F),
    functor(F,FN,_),
    !. 				% green

% Well-formed formula

wff([]).
wff([H|T]) :- 
    unvalued_uF(H) -> 
	format(string(Out),'~p has unvalued uF: derivation crash',[H]), 
	output(Out),
	fail 
    ;   wff(T).

%% Conceptual Intentional (CI) interface section

% check all CPs headed by interrogative C
ci(set(List)) :-
    !,				% red
    ciL(List).
ci(pm([X,Y])) :-
    !,				% red
    ci(X), ci(Y).	% pair merge
ci(CP) :-
    head(CP,C), label(C,c_Q),
%    uF(C,f(wh,_)), 
    !,				% red
    edge_list(CP,List),
    (ci_interrogative(List) ->
	 true
     ;  format(string(Out),
	       'Conceptual-Intensional (CI) Interface: ~p + edge ~p constitutes an ill-formed interrogative!',[C,List]),
	output_error(Out),
	assert(pending_step),	% increments if later steps
	fail
    ),
    complement(CP,Compl),
    ci(Compl).
ci(X) :-
    cs(X,Cs), ciL(Cs).

ciL([]).
ciL([C|Cs]) :- ci(C), ciL(Cs).

ci_interrogative(L) :-
    rmHeads(L,[]),
    !, 				% red
    output("CI: yes/no-question").

ci_interrogative(L) :- 
    forall(member(X,L), (wh_q(X) ; head(X))), 
    !,				% green
    output("CI: wh-question").

% exclamative: "if a non-wh-phrase appears as one of its specifiers"
ci_interrogative(L) :- 
    member(DP,L), label(DP,type(d)), \+ iF(DP,f(wh)),
    member(EP,L), exclamative(EP),
    !,				% green
    output("CI: exclamative").
    
% temporary
% [d [d1] [d [d2] .. ]] q what (a ...)
exclamative(DP) :- 
    iF(DP,f(wh)), complement(DP,DP2), label(DP2,what), uF(DP2,f('N',_)).

wh_q(DP) :- iF(DP,f(wh)), \+ exclamative(DP), !.
wh_q(PP) :- label(PP,type(p)), complement(PP,DP), wh_q(DP). % pied-piping

rmHeads([],[]).
rmHeads([X|L],List) :- (head(X) -> rmHeads(L,List) ; List = [X|L2], rmHeads(L,L2)).

edge_list(X,List) :-
    except_head(X,OL), butlast(OL,List).

except_head(X,L) :-
    (head_nonhead(X,H,NH) ->
	 L = [NH|Lp],
	 except_head(H,Lp)
     ;   L = []).

%% Grammar API

active(X) :- unvalued_uF(X).

edge_theta_mark(F,X) :-
    ((F == unvalued_iF(theta) ;
      F == internal(unvalued_iF(theta))) ->
	 theta_mark(X)
     ;   true).

% theta/expletive defined by a feature
nonTheta(X) :- expl(X).
theta(X) :- iF(X,f(theta,_)).
expl(X) :- iF(X,f(expl)).

% allow X to be theta-marked
% or pushed down to its subject (V [ Sbj [G Obj])?
% currently, assume [ Sbj [G Obj]] gets role to reassign

theta_mark(X) :-
    (iF(X,f(theta,V)) ->
	 var(V), V = marked,
	 printThetaMarked(X)
     ; true).

%%% NEW: special for c_rel, valuing is optional (if already valued)
value_case(V,Case,P,G) :-
    printProbegoal(P,G,values,Case,'case on'),
    ((f(P,id(ID)) -> V =.. [Case,ID] ; V = Case) ->
	 true
    ; (label(P,type(c)) -> true
      ;
       printValueCaseFail(G,V),
       fail)).

% in a list, there are no elements with unvalued uFs
no_uFL([]).
no_uFL([B|L]) :- boundaryMark(B), !, no_uFL(L).
no_uFL([X|L]) :- \+ unvalued_uF(X), no_uFL(L).

% edge for feature F true if has edge F feature and merged already (not head)
edge(X) :- unused_edge(X), \+ head(X).
edge(X,F) :- unused_edge(X,F), \+ head(X).
unused_edge(X,F) :- f(X,ef(F,V)), var(V).
unused_edge(X) :- f(X,ef(_,V)), var(V).
use_edge(X) :-
    f(X,ef(_,V)), var(V),
    !,				% red (could be multiple EFs)
    V = checked.
use_edge2(X) :- f(X,ef(_,checked)) -> true ; true.

% don't care how many times EF is used
has_edge(X,F) :- f(X,ef(F,_)), \+ head(X).

% assume edge feature can be
% 1. internal(F), restricted to internal merge, or
% 2. F (free)
externalMerge(F) :- var(F) -> true ; \+ F = internal(_).
% internalMerge(F) :- var(F) -> true ; F = internal(_).
internalMerge(_).

% Possibilities for F:
% variable: nop
% feature structure: type(name)
% type = iF, uF or f
% name = feature name
% Advanced:
% internal(variable): nop
% internal(feature structure): check

% X is goal: uCase(nom(ID)) = iF('T')
% ID is probe ID
checkfeature(X,_ID,iF('T')) :-	% which/who relative pronoun
    f_piedp(X,check('T')),
    !.				% green
checkfeature(X,ID,iF('T')) :- 
    uF_piedp(X,f(case,V,_)), V == nom(ID),
    \+ f(X,no_check('T')),	% block: DP *the boy called mary
    !.				% red
checkfeature(X,_ID,F) :- checkfeature(X,F).

checkfeature(X,F) :- 
    (var(F) ->
	 true 
     ;  (F = internal(F2) ->
	     (var(F2) ->
		  true 
	      ;  checkfeatureType(F2,X))
	 ;  checkfeatureType(F,X))).

%%% NEW: made iF(F) more flexible
checkfeatureType(unvalued_iF(F),X) :- iF(X,f(F,V)), var(V).
checkfeatureType(valued_iF(F),X) :- iF(X,f(F,V)), nonvar(V).
checkfeatureType(iF(F),X) :-
	iF(X,FC), functor(FC,Fn,_), (Fn == F -> true ; arg(1,FC,F)).
checkfeatureType(f(F),X) :- f(X,F).
checkfeatureType(uF(F),X) :- uF(X,f(F,_)).
checkfeatureType(uF(F),X) :- uF(X,f(F,_,_)).
checkfeatureType(unvalued_uF(F),X) :- unvalued_uF(X,F).
checkfeatureType(valued_uF(F),X) :- uF(X,f(F,V)), nonvar(V).
checkfeatureType(valued_uF(F),X) :- uF(X,f(F,V,_)), nonvar(V).

mkStream([],[]).
mkStream([C|Cs],[SO|SOs]) :-
    (C = [_|_] 
     -> mkStream(C,SO),
	mkStream(Cs,SOs)
     ;  (Cs = [C2|Csp]
	 -> (share_id(C,C2) % T c
	     -> ((type(C,Ct), xmit_uFs(C2,UFs,Ct)) 
		 -> mkSO(C,UFs,SO)
		 ;  mkSO(C,SO)),
		mkSO(C2,SO2), f(SO,id(V)), f(SO2,id(V)),
		SOs = [SO2|SOsp],
		mkStream(Csp,SOsp)
	     ;  mkSO(C,SO),
		mkStream(Cs,SOs))
	 ;  mkSO(C,SO),
	    mkStream(Cs,SOs)
	)
    ).

share_id(X,Y) :- type(X,'T'), type(Y,c).

% syntactic object so(LBL,uFs,iFs,Fs,SOs)
% Problem: assumes unambiguous C
mkSO(C,so(C,UFs,IFs,Fs,[])) :- 
    \+ features(C,UFs,IFs,Fs),
    !,				% red
    format(string(Out),'~w not found!',[C]), 
    output(Out),
    fail.
mkSO(C:Type,so(C,UFs,IFs,Fs,[])) :- 
    !, 				% red
    features(C:Type,UFs,IFs,Fs).
mkSO(C,so(C,UFs,IFs,Fs,[])) :- features(C,UFs,IFs,Fs).

% add uFs
mkSO(C,UFs2,so(C,UFs,IFs,Fs,[])) :- 
    (features(C,UFs1,IFs,Fs) ->
	 append(UFs1,UFs2,UFs)
     ;  format(string(Out),'~w not found!',[C]), 
	output(Out),
	fail).

%%% NEW: label can be conditional
% for printing
label_value(C,V) :-
	(atomic(C) ->
	 V = C
	;
	 C = c(C1,P),
	 (\+ \+ P -> V = C1 ; atom_concat(C1,'?',V))).

label_complex(so(C,_UFs,_IFs,_Fs,_Cs),C) :- \+ atomic(C).

label(pm([X,_]),C) :- label(X,C).
label(so(C,_UFs,_IFs,_Fs,_Cs),LB) :-
	var(LB),
	!,			% green
	(atomic(C) -> LB = C ; C = c(LB,_P)).
label(so(C,_UFs,_IFs,_Fs,_Cs),LB) :- 
	nonvar(LB),
	(LB = type(T) ->
	 (T = [_|_] -> member(U,T) ; U=T),
	 (atomic(C) -> C1 = C ; C = c(C1,_P)),
	 type(C1,U)
	;
	 LB == C).
% works with heads and non-heads
label_basetype(so(C,UFs,IFs,_Fs,_Cs),C0) :-
    (atomic(C) -> C1 = C ; C = c(C1,_P)),	
    (features(C1,UFs1,IFs1,_Fs1) ->
	 matchingUFsIFs(C1,UFs,UFs1,IFs,IFs1),
	 type(C1,C0)
     ;  features(C1:C0,UFs1,IFs1,_Fs1),
	matchingUFsIFs(C0,UFs,UFs1,IFs,IFs1)).
% doesn't work for T which inherits uv from C
matchingUFsIFs(C,UFs,UFs1,IFs,IFs1) :- \+ \+ (matchingUFs(C,UFs,UFs1), IFs = IFs1).
matchingUFs(C,UFs,UFs1) :-
    (UFs = UFs1 ->
	 true
     ; xmit_uFs(_,UFs1p,C),
       append(UFs1,UFs1p,UFs)).
uFs(pm([X,_]),UFs) :- uFs(X,UFs).
uFs(so(_,UFs,_,_,_),UFs).
iFs(pm([X,_]),IFs) :- iFs(X,IFs).
iFs(so(_,_,IFs,_,_),IFs).
uF(pm([X,_]),UF) :- uF(X,UF).
uF(so(_,UFs,_,_,_),UF) :- member(UF,UFs).
iF(pm([X,_]),IF) :- iF(X,IF).
iF(so(_,_,IFs,_,_),IF) :- member(IF,IFs).
fs(pm([X,_]),Fs) :- fs(X,Fs).
fs(so(_,_,_,Fs,_),Fs).
f(pm([X,_]),F) :- f(X,F).
f(so(_,_,_,Fs,_),F) :- member(F,Fs).
cs(pm([X,_]),Cs) :- cs(X,Cs).
cs(so(_,_,_,_,Cs),Cs).
c(pm([X,_]),C) :- c(X,C).
c(so(_,_,_,_,Cs),C) :- member(C,Cs).

% supports pied piping
uF_piedp(X,UF) :- uF(X,UF) -> true ; complement(X,Y), uF(Y,UF).
iF_piedp(X,IF) :- iF(X,IF) -> true ; complement(X,Y), iF(Y,IF).
f_piedp(X,F) :- f(X,F) -> true ; complement(X,Y), f(Y,F).

% order_cs(XY,[X,Y])
% X is head
%   parameter head(first):
%     [H,Y]
%     [Y,H] if Y has feature attach(left)
%   otherwise:
%     [Y,H]
%     [H,Y] if Y has feature attach(right)
% override:
%    head(first(H)) type on H
%    head(final(H))

order_cs(so(_,_,_,_,Cs),[H,Y]) :-
    head(H),
    !,				% red
    label_basetype(H,H0),	% needed: e.g. what:d, what:n, o/w double solution
    (parameter(head(first(H0))) ->
	 attachLeft(H,Y,Cs)
     ; (parameter(head(final(H0))) ->
	    attachRight(H,Y,Cs)
	; (parameter(head(first)) ->
	       attachLeft(H,Y,Cs)
	  ; attachRight(H,Y,Cs)))),
    !.				% red: is label_basetype ambiguous for v~?

order_cs(so(_,_,_,_,[Y,CH]),[CH,Y]). % 2nd merge always to the left

attachLeft(H,Y,Cs) :-
    (f(Y,attach(left)) ->
	 printAttach(Y,left),
	 Cs = [Y,H]
     ;  Cs = [H,Y]).

attachRight(H,Y,Cs) :-
    (f(Y,attach(right)) ->
	 printAttach(Y,right),
	 Cs = [H,Y]
     ;  Cs = [Y,H]).

head(so(_,_,_,_,[])).
syntactic_head(X) :- head(X) -> true ; f(X,morphology).

headInitial(Type) :-
    (parameter(head(first(Type))) ->
	 true
     ; \+ parameter(head(final(Type))),
       parameter(head(first))).

% X is adjunct, Y is SO
pair_merge(X,Y,pm([Y,X])) :-
    printPairMerge(X,Y).

% [C Edge [C ..]] recurse on [C ..]
% [C [C  ] Compl]
complement(so(C,_,_,_,[so(C,_,_,_,[]),Compl]),Compl) :- !. % red
complement(so(C,_,_,_,[_Edge,SO]),Compl) :- label(SO,C), complement(SO,Compl).
head(X,X) :- head(X), !. % red
head(X,H) :- label(X,C), cs(X,L), member(Y,L), label(Y,C), head(Y,H).

% heads in phrase X (left edge heads + head of phrase)
heads(X,Heads) :- heads(X,[],Heads).

heads(X,Hds,[X|Hds]) :- head(X), !. 	% red
heads(X,Hds,Hdsp) :- 
    label(X,C), cs(X,L),
    headsL(L,C,Hds,Hdsp).

headsL([],_C,Hds,Hds).
headsL([X|L],C,Hds,Hdsp) :-
    (head(X) ->
	 (label(X,C) ->
	      Hdsp = [X|Hds]
	  ;  headsL(L,_,[X|Hds],Hdsp))
     ;  (label(X,C) ->
	     heads(X,Hds,Hdsp) 
	 ;  headsL(L,C,Hds,Hdsp))).

head_nonhead(X,H,NH) :-
    label(X,C), cs(X,L), pick(H,L,[NH]), label(H,C), 
    !. % green

in_edge(X,Y) :- label(X,C), in_edge(X,C,Y).

in_edge(X,C,_) :-
    head(X), label(X,C), !, fail.
in_edge(X,C,Y) :-
    label(X,C), cs(X,L), in_edgeL(L,C,Y).

in_edgeL([X|L],C,Y) :-
    (X = Y ->
	 true
     ;  (head(X) ->
		\+ label(X,C), X = Y
	 ;  (label(X,C) ->
		 in_edge(X,C,Y) 
	     ;  in_edgeL(L,C,Y)))).

feature_name(F,FN) :- arg(1,F,FN).

set_occ(so(_,_,_,Fs,[])) :- member(occ(X),Fs) -> X =1 ; true.

% new_occ(X,X')
% if occ(_) in X, and unvalued, X' has fresh occ(_)
% if occ(_) in X, already valued,
%   X' copies if set_occ_flag false,
%   X' fresh occ(_) if set_occ_flag true

new_occ(set(List),set(Listp)) :-
	!,			% red
	new_occL(List,Listp).
new_occ(pm([X,Y]),pm([Xp,Yp])) :-
    !, % red
    new_occ(X,Xp),
    new_occ(Y,Yp).
new_occ(so(C,UF,IF,F,[]),so(C,UF,IF,Fp,[])) :-
    !,				% red
    inherit_Fs_newOcc(Fp,F).
new_occ(so(C,UF,IF,F,Cs),so(C,UF,IF,F,Csp)) :-
    new_occL(Cs,Csp).

% if occ(_), set occ(1) for old if not already set
% if occ(_) already set, don't create new occ, keep occ(1)
% if occ(_) not present, don't create one unless occ_new_flag is set

inherit_Fs_newOcc(Fp,F) :-
    (append(Left,[occ(V)|Right],F) ->
	 (var(V) ->
	      V = 1, 		% set old occ
	      append(Left,[occ(_)|Right],Fp),
	      !			% green
	  ; (occ_new_flag ->
		 append(Left,[occ(_)|Right],Fp)
	     ;  Fp = F))
     ;   Fp = F).

new_occL([],[]).
new_occL([X|L],[Y|Lp]) :-
    new_occ(X,Y),
    new_occL(L,Lp).

% two heads are the same except for substream(_) feature
% Note 1: substream(_) is introduced by add_substreamF/2 to
% identify stack items formed in a substream
% Note 2: used in subC checking

sameExceptSSF(X,Y) :- \+ \+ sameExceptSSF1(X,Y).

sameExceptSSF1(so(LB,UF,IF,FF1,CS),so(LB,UF,IF,FF2,CS)) :-
    rmSSF(FF1,FF),
    rmSSF(FF2,FF).

rmSSF([],[]).
rmSSF([X|L],Lp) :- (X = substream(_) -> rmSSF(L,Lp) ; Lp = [X|L2], rmSSF(L,L2)).
		  
% go through SO, set all head occ variables
set_occs(X) :-
    (head(X) ->
	 set_occ(X)
     ;   cs(X,L),
	 set_occsL(L)).

set_occsL([]).
set_occsL([X|L]) :-
    set_occs(X),
    set_occsL(L).
	
% like new_occ(X,Xp) except always fresh occ for occ heads
% Note: used for stack subconstituents
new_occ_override(pm([X,Y]),pm([Xp,Yp])) :-
    !, % red
    new_occ_override(X,Xp),
    new_occ_override(Y,Yp).
new_occ_override(so(C,UF,IF,F,[]),so(C,UF,IF,Fp,[])) :-
    !,				% red
    inherit_Fs_newOcc_override(Fp,F).
new_occ_override(so(C,UF,IF,F,Cs),so(C,UF,IF,F,Csp)) :-
    new_occL_override(Cs,Csp).

inherit_Fs_newOcc_override(Fp,F) :-
    (append(Left,[occ(_)|Right],F) ->
	 append(Left,[occ(_)|Right],Fp),
	 !			% green
     ;   Fp = F).

new_occL_override([],[]).
new_occL_override([X|L],[Y|Lp]) :-
    new_occ_override(X,Y),
    new_occL_override(L,Lp).

% mark_occ(_).
% recursion for multiple occ features
mark_occ(X) :-
    ((f(X,occ(V)), var(V)) ->
	 V = 1,
	 mark_occ(X)
     ;   true).

% unvalued_uF(X) true if X has unvalued uninterpretable features
% unvalued_uF(X,Fn) true if X has unvalued uninterpretable feature Fn
% unvalued_uF(X,ID,Fn) true if X (with ID) has unvalued uninterpretable feature Fn
% Fn = feature name

unvalued_uF(X) :- 
    uFs(X,UFs), member(F,UFs), arg(2,F,V), unvalued(V), 
    !.				% green

unvalued_uF(X,Fn) :- 
    uFs(X,UFs), 
    (var(Fn) ->
	 % iterate..
	 member(F,UFs), arg(2,F,V), unvalued(V), arg(1,F,Fn)
     ;
     member(F,UFs), arg(2,F,V), unvalued(V), arg(1,F,Fn), !).

unvalued_uF(X,ID,Fn) :- 
    f(X,id(ID)), uFs(X,UFs), 
    member(F,UFs), arg(2,F,V), unvalued(V), arg(1,F,Fn), 
    !.				% red

value_uF(X,Fn) :- 
    uFs(X,UFs), member(F,UFs), arg(1,F,Fn), arg(2,F,1), 
    !.				% green

% f(phi,V) V a variable
% f(phi,[V,x,x]) V (person) subfeature a variable
% f(case,null(V),_) V a variable  
% f(case,opt(V,C),_) - NOT unvalued  

unvalued(V) :-
    (var(V) ->
	 true
     ;	((member(SubV,V), var(SubV)) ->
	     true
	 ;   V = null(V2), var(V2))).

% feature sharing
% if not both X and Y have feature F, quietly continues
share_uFs(X,Y,L) :- share_uFs1(L,X,Y).
share_uFs1([],_,_).
share_uFs1([F|Fs],X,Y) :- 
    ((uFs(X,XFs), uFs(Y,YFs), member(F,XFs), member(F,YFs)) -> true ; true),
    !,				% green
    share_uFs1(Fs,X,Y).

share_iFs(X,Y,L) :- share_iFs1(L,X,Y).
share_iFs1([],_,_).
share_iFs1([F|Fs],X,Y) :- 
    iFs(X,XFs), iFs(Y,YFs), member(F,XFs), member(F,YFs),
    !,				% green
    share_iFs1(Fs,X,Y).

% share if such features exists
share_iFs_opt(X,Y,L) :- share_iFs1_opt(L,X,Y).
share_iFs1_opt([],_,_).
share_iFs1_opt([F|Fs],X,Y) :- 
	((iFs(X,XFs), iFs(Y,YFs), member(F,XFs), member(F,YFs)) ->
	 true ; true),
	share_iFs1_opt(Fs,X,Y).

share_Fs(X,Y,L) :- share_Fs1(L,X,Y).
share_Fs1([],_,_).
share_Fs1([F|Fs],X,Y) :- 
    fs(X,XFs), fs(Y,YFs), member(F,XFs), member(F,YFs),
    !,				% green
    share_Fs1(Fs,X,Y).

% X projects if
% (a) X is a head without a noProject feature, or
% (b) X is a non-head with an unused EF
% Note: set does not project
projects([X,_]) :- projects(X).
projects(so(_,_,_,Fs,Cs)) :- 
    (Cs == [] ->
          \+ member(noProject,Fs)
     ; member(ef(_,V),Fs), var(V)).

projectingHead(X) :- syntactic_head(X), \+ f(X,noProject).

%% Search operations
%%   section cincorporates notion of phase and local extent

is_phase(X) :- f(X,phase(_)), \+ projects(X), !. % green

% complete phase:
% 1. finished projecting
% 2. no uF on head or in edge (parameter: le(pc))
% 3. not marked as completed (yet)
is_complete_phase(X) :-
    \+ projects(X),
    is_uncompleted_phase(X),
    (le(pc) -> no_unvalued_uF_edgehead(X) ; true).

no_unvalued_uF_edgehead(X) :-
    \+ unvalued_uF(X),
    edge_list(X,L),
    forall(member(E,L),(\+ unvalued_uF(E))).

is_phase_criterion(X) :-
    le(on),
    !,				% red
    is_phase(X).
is_phase_criterion(X) :-
    le(pc),
    !,				% red
    is_complete_phase(X).
is_phase_criterion(_) :-
    le(off),
    fail.

% make sure phase is subject to completion once only
is_uncompleted_phase(X) :- f(X,phase(V)), var(V).
completed_phase(X) :- f(X,phase(V)), nonvar(V).

% 1 version skips the machine parameterization
mark_completed_phase1(X) :- f(X,phase(V)), var(V), !, V = c.

% parameterized by machine parameter le(pc)
mark_completed_phase(X,ST,STp) :-
    ((le(pc) ; le(on)) ->
     labelingFailureCheck(ST),
     adjustStackAtLE(ST,STp)
    ;
     STp = ST),
    mark_completed_phase1(X).

%%% NEW: check stack
labelingFailureCheck([]).
labelingFailureCheck([H|T]) :-
	(H == b ->
	 true
	;
	 (H = c(C,Goal) ->
	  (\+ \+ Goal ->
	   labelingFailureCheck(T)
	  ;
	   printLabelingFailure(C),
	   fail)
	 ;
	  labelingFailureCheck(T))).

% self: feature release(_)
% [self[self][he[he][d[d][John]]]] , [d[d][John]] can only be released once

% search parameter:
%   so: search SO for X with unvalued uF and RC
%       returns list [..(X,RC)..]
%   stk: search stack for X with inactive(unvalued uF)
%       returns list of [..X..]
%   off: search_release fails
% 
% RC is release control SO: e.g. self, single release only
% RC is unbound: multiple releases possible

search_release(SO,ST,ST,L) :-
    search(so),
    !, 				% red
    search_SO_uFs(SO,L).
search_release(_,ST,STp,Xs) :-
    search(stk),
    !,				% red
    search_stack_iuFs(ST,Xs,STp).

% Note: Xs preserves stack order
search_stack_iuFs(L,Xs,Lp) :- search_stack_iuFs(L,[],Xs,Lp).

search_stack_iuFs(L,Xs,Xsp,Lp) :-
    (search_stack_iuF(L,X,L1) ->
	 Xs2 = [X|Xs],
	 search_stack_iuFs(L1,Xs2,Xsp,Lp)
     ;   Xsp = Xs,
	 Lp = L).

% search_stack_iuF(Stk,X,Stk')
% pied-piping support
search_stack_iuF(L,Z,Lp) :-
    append(Left,[X|Right],L),
    (unvalued_uF_piedp(X) ->
     Z = X,
     printStackSearchUF(X),
     append(Left,Right,Lp)
    ;
     %% (19a) Pictures of himself1 are hard for John1 to like
     %% DC [pro pictures of self he john]
     %% [pro[pro][d!case[d][pictures[pictures][of[of][self[self][he[he][d[d][john]]]]]]]]
     %% inactive([d!case[d][pictures[pictures][of[of][self[self][he[he][d[d][john]]]]]]])
     functor(X,inactive,1), arg(1,X,Y),
     unvalued_uF_piedp(Y),
				% Why this hack? What for?
				%	new_occ(Y,Yp),		% spellout hack
				%        Z = Yp,
     Z = Y,
     printStackSearchInactive(X),
     append(Left,Right,Lp)),
    !.				% red

% unvalued_uF with pied piping support
% looks at complement of phrase
% for prepositions only atm
% NEW: for [d_rel d_rel N!D] too
unvalued_uF_piedp(X) :-
	unvalued_uF(X),
	!.			% red
unvalued_uF_piedp(X) :-
%	label(X,type(p)),
	(label(X,type(p)) -> true ; label(X,d_rel)),
	complement(X,Y),
	unvalued_uF(Y).

search_SO_uFs(X,L) :-
    findall((Y,Z),search_SO_uF(X,Y,Z),L).

% search_SO_uF(X,Y,Z) true if
% Y subconstituent of X, and
% Y has unvalued uF, and
% Z is release control that dominates Y
% if Z is bound, i.e. exists, Y is single release
% if Z is not bound, Y is recursive release
search_SO_uF(X,Y,Z) :-
    (\+ (is_phase(X), f(X,substream(_)))),
    (f(X,release(V)) -> var(V), Z=X ; true),
    (edge_list(X,L), member(W,L) ; complement(X,W)),
    search_SO1(W,Y,Z).

% 1. block if released, i.e has release(_) feature and used,
%    cannot be used again.
% 2. block if it came from a substream phase, i.e.
%    phase boundary with has feature substream(_) 
% 3. has unvalued uF, candidate
% 4. has feature release(_), not released.

search_SO1(X,_,_) :-
    (released(X) ; is_phase(X), f(X,substream(_))),
    !,				% red
    fail.
search_SO1(X,X,_) :-
    unvalued_uF(X),
    !.				% red
search_SO1(X,Y,Z) :-
    (f(X,release(V)) ->
	 var(V),
	 Z = X
     ;   true),
    c(X,W),
    search_SO1(W,Y,Z).

% release_to_stackL(RL,LE,ST,ST')
% LE = local extent triggering release
% RL = release list [..(X,Z)..] (SO search) or [..X..] (stack search)
% X = element to be released (original) onto stack ST
% Xp2 = copy of X to be stored on stack
% Z only applies for SO search (not stack search)
%   if nonvar, Z dominates X (and is marked to stop subsequent searches)
%
% Occ: controls spellout (first occ gets spelled out)
%	old copy of X gets occ(1), i.e. will never be spelled out
%	stacked copy of X gets fresh occ(_)
% Note: occ(1) is used to prevent spellout of subsequent movement
% of original element ABOVE stacked, e.g. Pictures of himself worry John t

releaseL_to_stack(RL,LE,ST,STp) :-
    mark_completed_phase(LE,ST,ST1), % may adjust the stack if drop_stack
%    mark_released(LE),
    releaseL_to_stackL(RL,LE,ST1,STp).

releaseL_to_stackL([],_LE,ST,ST).
releaseL_to_stackL([RL|L],LE,ST,STp) :-
    (RL = (X,Z) ->
	 release_to_stack(X,Z,LE,ST,ST1)
     ;  release_to_stack(RL,_,LE,ST,ST1)),
    releaseL_to_stackL(L,LE,ST1,STp).

release_to_stack(X,Z,LE,ST,STp) :-
    mark_releasedNF(Z),
%    new_occ(X,Xp), % X' fresh, X marked (no spellout)
%    mark_occ(X),
%    add_wrapper(LE,Xp,Xp2),	% add once(X') if search(stk)
    add_wrapper(LE,X,Xp2),	% add once(X') if search(stk)
    STp = [Xp2|ST],
%    printRelease(Xp).
    printRelease(X).

% Shift phase boundaries on stack
% 1. Mark the boundary: push b
% 2. Drop off items that are 2 boundaries down
% Machine parameter:
%  le(on/off/pc)	only mark using le if le(on) or le(pc)
%  drop_stack		on: drop elements 2 LEs down, off: no drop
% Notes: called by
% 10:c 2m/none 		(built CP)
% 7a:v/a/d 2m int	(built vP/DP/AP)
% 7b:d 2m (stream)	(built DP)
% 7b:v 2m (stream)	(built vP)
% 1b:D+cN		(DP - complex NP)

adjustStackAtLE(ST,ST) :-
    le(off),
    !.				% red
adjustStackAtLE(ST,STp) :-
    (le(on) ; le(pc)),
    !,				% red
    markPhaseBoundary(ST,ST1),
    (drop_stack ->
	 (drop2LE(ST1,Dropped,STp) ->
	      printDropped(Dropped)
	  ;   STp = ST1)
     ;   STp = ST1).

% have stuff to drop off the end of the stack (after 2nd LE)
% could drop just the last b
drop2LE(L,Dropped,Lp) :-
    boundaryMark(B),
    L = [B|L1],
    append(Left,[B|Dropped],L1),
    !,				% red
    Lp = [B|Left].

%% Merge

% merge(X,Y,XY,Hd)
% Hd = head of phrase (labeler)
% labeling rules for XY
% 1. X with EF (not first merge), X labels XY (can reuse edge)
% 1. Y with EF (not first merge), Y labels XY (use edge)
% 2. projecting head X merge non-head Y, X labels XY
% 3. projecting head Y merge non-head X, Y labels XY
% 4. head X merge head Y, (a) either X or (b) Y with uF labels XY
% 5. X, Y non-heads, assume X is existing SO, X labels
% Possible issue:
%   substream -> main stream means head ... how to signal this?
%%% NEW: override by specifying head parameter
%%% NEW: XP YP merge permitted
merge(X,Y,XY,Param) :-	% [X,Y] LR order: headfinal
	Param == headfinal,
	!,
	inherit_from(XY,Y),
	cs(XY,[X,Y]),
	printMerge(X,Y,Y,'head-final specified'),
	!.			% red
merge(X,Y,XY,Param) :-	% [X,Y] LR order: headinitial
	Param == headinitial,
	!,
	inherit_from(XY,X),
	cs(XY,[X,Y]),
	printMerge(X,Y,X,'head-initial specified'),
	!.			% red

merge(X,Y,YX,X) :- % X has F-edge feature, merge to left
	has_edge(X,_),		% can reuse
	!,
	inherit_from(YX,X), cs(YX,[Y,X]),
%	use_edge(YX),
	printMerge(X,Y,X,'edge feature').
merge(X,Y,XY,X) :- % head X, non-head Y
	projectingHead(X), \+ syntactic_head(Y),
	!,	
	printMerge(X,Y,X,'syntactic head merging with a non-head'),
	inherit_from(XY,X,Y), order_cs(XY,[X,Y]).
merge(X,Y,XY,Y) :- 	% non-head X, head Y
	\+ syntactic_head(X), projectingHead(Y),
	!,
	printMerge(X,Y,Y,'syntactic head merging with a non-head'),
	inherit_from(XY,Y,X), order_cs(XY,[Y,X]).
% X and Y heads:
% (1) X uF Y => [X X Y] (2) X Y uF => [Y X Y] (3) X Y => [Y Y X] (default)
merge(X,Y,XY,H) :-	% head X, head Y
	syntactic_head(X), syntactic_head(Y),
	!,	
	(unvalued_uF(X) ->
	     printMerge(X,Y,X,'syntactic head with an unvalued uF'),
	     inherit_from(XY,X,Y), order_cs(XY,[X,Y]), H = X
	 ;  (unvalued_uF(Y) ->
		 printMerge(X,Y,Y,'syntactic head with an unvalued uF'),
		 inherit_from(XY,Y,X), order_cs(XY,[Y,X]), H = Y
	     ;  ((projects(X), \+ projects(Y)) ->
		     printMerge(X,Y,X,'only projecting head'),
		     inherit_from(XY,X,Y), order_cs(XY,[X,Y]), H = X
		 ;  ((\+ projects(X), projects(Y)) ->
			 printMerge(X,Y,Y,'syntactic head with an unvalued uF'),
			 inherit_from(XY,Y,X), order_cs(XY,[Y,X]), H = Y
		     ;  projects(X), projects(Y),
			printMerge(X,Y,X,'leftmost by default'),
			inherit_from(XY,X,Y), order_cs(XY,[X,Y]), H = X)))).
merge(X,Y,XY,X) :- % non-head X, non-head Y, assume X labels
	\+ syntactic_head(X), \+ syntactic_head(Y), projects(X),
	!,
	inherit_from(XY,X),
	order_cs(XY,[X,Y]),
	printMerge(X,Y,X,'existing SO labels').
%%% NEW
%%% neither argument projects, neither is a head
%%% catch-all case: e.g. ext Merge XP adjunct that must be moved from SO
%%% put it on the right (assume rightwards adjunction)
merge_special(XP,SO,SOXP) :-
	inherit_from_cond(SOXP,SO,XP), % conditional labeling
	cs(SOXP,[SO,XP]),	% fool into putting XP on the right
	printMergeCond(XP,SO,SO,'XP-YP general case').

% inherit_from(X,Y) X inherits all features from Y, except occ feature
% two non-heads case
inherit_from(so(LB,U,I,Fp,_),so(LB,U,I,F,_)) :- 
    inherit_Fs_noOcc(Fp,F).

% inherit_from(X,Y,Z) X inherits features from Y plus iFs and Fs from Z
% if Y is D and Z is N (Z could be a non-head), or
% if Y is A and Z is N

inherit_from(so(LB,U,I,F,_),so(LB,U,I1,F1,_),Y) :-
    (inherit_iFs(LB,Y) ->
	 iFs(Y,I2), fs(Y,F2),
	 inherit_Fs_noOcc(F2p,F2),
	 printInherit(Y,I2,F2),
	 merge_Fs(I1,I2,I),
	 inherit_Fs_noOcc(F1p,F1),
	 append(F1p,F2p,F)
     ;  I = I1,
	inherit_Fs_noOcc(F,F1)).

inherit_from_cond(so(c(LB,nonvar(Occ)),U,I,F,_),X,Y) :-
	X = so(LB,U,I1,F1,_),
	head(Y,H),
	f(H,occ(Occ)),
	(inherit_iFs(LB,Y) ->
	 iFs(Y,I2), fs(Y,F2),
	 inherit_Fs_noOcc(F2p,F2),
	 printInherit(Y,I2,F2),
	 merge_Fs(I1,I2,I),
	 inherit_Fs_noOcc(F1p,F1),
	 append(F1p,F2p,F)
	;
	 I = I1,
	 inherit_Fs_noOcc(F,F1)).

inherit_iFs(LB1,Y) :-
    type(LB1,T1), label(Y,LB2), type(LB2,T2),
    inherit_matrix(T1,T2).
    
inherit_matrix(d,d).		% what a, how many
inherit_matrix(d,n).
inherit_matrix(d,a).
inherit_matrix(a,n).

% occ is removed
inherit_Fs_noOcc(Fp,F) :-
    (append(Left,[occ(_)|Right],F) ->
	 append(Left,Right,Fp),
	 !			% green
     ;   Fp = F).

merge_Fs([],Fs,Fs).
merge_Fs([F|Fs1],Fs2,Fs) :-
    (f_in(F,Fs2,F2) ->
	 F = F2,
	 merge_Fs(Fs1,Fs2,Fs)
     ;  Fs = [F|Fsp],
	merge_Fs(Fs1,Fs2,Fsp)).

f_in(f(F),L,f(F)) :- memberchk(f(F),L).
f_in(f(F,_),L,f(F,V)) :- memberchk(f(F,V),L).
f_in(f(F,_,_),L,f(F,V,O)) :- memberchk(f(F,V,O),L).

%% stack operations

% X is two boundaries (b) down in L
twoLEdown_uF(L,X) :-
    boundaryMark(B),
    L = [B|L1],
    append(Left,[B|TwoDown],L1),
    member(X,TwoDown),
    unvalued_uF(X),
    \+ member_nounify(X,Left).

% no unify member
member_nounify(X,[Y|_]) :- X == Y.
member_nounify(X,[_|L]) :- member_nounify(X,L).

% push if uF present
% handles inactive(_) wrapper
% handles pm no_pied_pipe
opush(pm([X,Y]),S,Sp) :-	% doesn't do inactive(pm(_))
    !, 				% red
    (f(Y,no_pied_pipe) ->
	 printPushNoPiedPipe(Y),
	 opush(X,S,Sp)
     ;  opush_core(pm([X,Y]),S,Sp)).
opush(IX,S,Sp) :-
    functor(IX,inactive,1), arg(1,IX,X),
    !,				% red
    (unvalued_uF(X) ->
	 (checkSubconstituent(X,S,S1) ->
	      push2(inactive(X),S1,Sp),
	      printPush(inactive(X))
	  ;  push2(inactive(X),S,Sp),
	     printPush(X))
     ;   S = Sp).
opush(X,S,Sp) :- opush_core(X,S,Sp).

% if no uF, no need to stack
opush_core(X,S,Sp) :-
    (unvalued_uF(X) ->
	 (checkSubconstituent(X,S,S1) ->
	      push2(X,S1,Sp)
	  ;  push2(X,S,Sp)),
	 printPush(X)
     ;  S = Sp).

% list to push, takes only the head
pushL_hd([],S,S).
pushL_hd([X|L],S,Sp) :- head(X,H), push(H,S,S1), pushL_hd(L,S1,Sp).

push(X,S,Sp) :-
    (checkSubconstituent(X,S,S1) -> 		% S1 = delete sub from S
	 push2(X,S1,Sp)
     ;  push2(X,S,Sp)),
    printPush(X).

% push X onto stack S, subconstituent of X occurs in S,
% mark sub on S,
% new occ for sub on S
% X should be partially marked

% stack_subc parameter: true, mark and substitute on stack, false: just push
push2(IX,S,[IX|S2]) :-
    stack_subc,			% parameter
    functor(IX,inactive,1), arg(1,IX,X),
    append(Left,[Y|Right],S),
    !, 				% red
    ((functor(Y,once,1), arg(1,Y,Y2)) ->
	 subC(Y2,X,_),
	 new_occ(Y2,Y2p),
	 !,
	 append(Left,[once(Y2p)|Right],S2)
     ;  ((functor(Y,inactive,1), arg(1,Y,Y2)) ->
	     subC(Y2,X,_),
	     new_occ(Y2,Y2p),
	     !,
	     append(Left,[inactive(Y2p)|Right],S2)
	 ;  subC(Y,X,_),
	    !,
	    new_occ(Y,Yp),
	    append(Left,[Yp|Right],S2))).
push2(X,S,[X|S2]) :-
    stack_subc,			% parameter
    append(Left,[Y|Right],S),
    ((functor(Y,once,1), arg(1,Y,Y2)) ->
	 subC(Y2,X,_),
	 new_occ(Y2,Y2p),
	 !,
	 append(Left,[once(Y2p)|Right],S2)
     ;  ((functor(Y,inactive,1), arg(1,Y,Y2)) ->
	     subC(Y2,X,_),
	     new_occ(Y2,Y2p),
	     !,
	     append(Left,[inactive(Y2p)|Right],S2)
	 ;  subC(Y,X,_),
	    !,
	    new_occ(Y,Yp),
	    append(Left,[Yp|Right],S2))).
push2(X,S,[X|S]).
        
frontStack(X,ST,STp) :-
    ((append(L,[Xc|R],ST), L \== [], Xc == X) ->
	 append([Xc|L],R,STp),
	 printStackFronting(Xc)
     ;  STp = ST).
    
markPhaseBoundary(ST,[B|ST]) :- boundaryMark(B), printStackLE.

% pop and repeat for those with no uF present
opopR([],[]).
opopR(S,Sp) :- 
    S = [X|T], 
    (unvalued_uF(X) -> Sp = S ; printPop(X), opopR(T,Sp)).

% pop if no uF present
% no printing
opop1(S,Sp) :- 
    S == [] ->
	Sp = S
    ;  S = [X|T], 
       (unvalued_uF(X) -> Sp = S ; Sp = T).

opop1f([],[]) :- !.			     % green
opop1f(S,S) :- S = [X|_], unvalued_uF(X), !. % red
opop1f([_|S],S).

% don't bind occ(_)
wrapped_eq(X,Y) :- Y == X -> true ; \+ \+ (Y = once(X)).
	    
repush(X,S,Sp) :-		% already front
    S = [Y|L],
    wrapped_eq(X,Y),
    !,				% red
    Sp = [X|L].
repush(X,S,Sp) :-		% already front
    boundaryMark(B),
    S = [B,Y|L],
    wrapped_eq(X,Y),
    !,				% red
    Sp = [B,X|L].
repush(X,S,Sp) :-		% front X
    append(Left,[Y|Right],S),
    Left \== [],
    wrapped_eq(X,Y),
    append([X|Left],Right,Sp),
    !,
    printFront(X).

% handles inactive wrap as well
replaceInStack(X,Xnew,ST,STp) :-
    (append(Left,[X|Right],ST) ->
	 append(Left,[Xnew|Right],STp)
     ;  append(Left,[inactive(X)|Right],ST),
	append(Left,[inactive(Xnew)|Right],STp)),
    !.				% green

% inactive needed?
replaceInGoals([],_,[]).
replaceInGoals([G|Gs],L,[X|Gsp]) :-
    member(X,L),
    (\+ \+ G = X),
    !,				% red
    replaceInGoals(Gs,L,Gsp).
replaceInGoals([G|Gs],L,[G|Gsp]) :-
    replaceInGoals(Gs,L,Gsp).

% Top of stack (TOS)
% respects inactive(_) wrapper
% stack_access parameter:
%  local: can't look past b
%  global: skips b markers
% digs into once(_) wrapper
tos_ncp([TOP|L],X) :-
    (functor(TOP,inactive,1) ->
	 tos_ncp(L,X)
     ;  (boundaryMark(TOP) ->
	     stack_access(global),
	     tos_ncp(L,X)
	 ;  (functor(TOP,once,1) ->
		 arg(1,TOP,X)
	     ;  X = TOP))).

% tos(ST,X,ST') returns TOS X
% old copy of X on stack is marked (through) - side effect.
% new X (with fresh occ) is returned
% old X in stack is also replaced with new X.

tos([TOP|L],X,[TOP|Lp]) :-
    (functor(TOP,inactive,1) ->
	 tos(L,X,Lp)
     ;  (boundaryMark(TOP) ->
	     !,
	     stack_access(global),
	     tos(L,X,Lp))).
tos([once(X)|L],Xnew,[once(Xnew)|L]) :-
    new_occ(X,Xnew),
    !.				% red
tos([X|L],Xnew,[Xnew|L]) :-
    new_occ(X,Xnew).

tos2([TOP|L],X,Y,Lp) :-
	(functor(TOP,inactive,1) ->
	 tos2(L,X,Y,Lp)
	;
	 \+ boundaryMark(TOP),
	 L = [Y|Lp],
	 \+ boundaryMark(Y),
	 X = TOP).

% relativized TOS
% rel_tos(STK,X,F,STK') finds first stack element with F
% respects inactive(_) wrapper
% no pied-piping support
% stack_access parameter:
%  local: can't look past b
%  global: skips b markers
% spellout: new occ, old occ - results in modified STK
% if subconstituent of found stack element is also on the stack
% it will be occ valued
% problem: should be modified with new occ too!
% because it could be extracted later

rel_tos(ST,X,F,STp) :-
    rel_tos1(ST,X,F,Xold,Old,New),
    updateStack(ST,Xold,X,Old,New,STp).

% updateStack(ST,Xold,X,Old,New,ST')
% look for both subconstituents of Xold on the stack ST
% and stack items containing Xold
% update them to reflect the (new) X
% special case: exact match to Old on stack -> New
updateStack([],_Xold,_X,_Old,_New,[]).
updateStack([Y|L],Xold,X,Old,New,[New|Lp]) :-
    Y == Old,			% truly identical
    !,				% red
    updateStack(L,Xold,X,Old,New,Lp).
updateStack([Y|L],Xold,X,Old,New,[Yp|Lp]) :-
    stack_update_larger,	  % machine parameter
    has_subC_update(Y,Xold,X,Yp), % Y has Xold as a subC
    !,				% red
    updateStack(L,Xold,X,Old,New,Lp).
updateStack([Y|L],Xold,X,Old,New,[Yp|Lp]) :- 
    subC_update(Xold,X,Y,Yp),	% Xold has Y as a subC
    !,				% red
    updateStack(L,Xold,X,Old,New,Lp).
updateStack([Y|L],Xold,X,Old,New,[Y|Lp]) :- % X and Y unrelated
    updateStack(L,Xold,X,Old,New,Lp).

% has_subC_update(Y,X,X',Y') Y has subC X, Y' is Y with X updated to X'
% Y is a stack item
has_subC_update(inactive(Y),X,Xp,inactive(Yp)) :-
    !,				% red
    has_subC_update1(Y,X,Xp,Yp).
has_subC_update(once(Y),X,Xp,once(Yp)) :-
    !,				% red
    has_subC_update1(Y,X,Xp,Yp).
has_subC_update(Y,X,Xp,Yp) :-
    has_subC_update1(Y,X,Xp,Yp).

has_subC_update1(Y,X,Xp,Xp) :-
    sameExceptSSF(Y,X),
    !.				% red
has_subC_update1(so(LB,UFs,IFs,Fs,L),X,Xp,so(LB,UFs,IFs,Fs,Lp)) :-
    has_subC_updateL(L,X,Xp,Lp).

has_subC_updateL([Y|L],X,Xp,[Yp|L]) :-
    has_subC_update1(Y,X,Xp,Yp),
    !.				% red
has_subC_updateL([Y|L],X,Xp,[Y|Lp]) :-
    has_subC_updateL(L,X,Xp,Lp).

% subC_update(X,X',Y,Y') Y is a subC of X, Y' is Y updated given X'
% Y is a stack item
subC_update(X,Xp,inactive(Y),inactive(Yp)) :-
    !,				% red
    subC_update1(X,Xp,Y,Yp).
subC_update(X,Xp,once(Y),once(Yp)) :-
    !,				% red
    subC_update1(X,Xp,Y,Yp).
subC_update(X,Xp,Y,Yp) :- subC_update1(X,Xp,Y,Yp).

subC_update1(X,Xp,Y,Xp) :-
    sameExceptSSF(X,Y),
    !.				% red
subC_update1(X,Xp,Y,Yp) :-
    cs(X,L), cs(Xp,Lp),
    member_parallel(Z,L,Zp,Lp),
    subC_update1(Z,Zp,Y,Yp).

member_parallel(X,[X|_],Y,[Y|_]).
member_parallel(X,[_|L],Y,[_|Lp]) :- member_parallel(X,L,Y,Lp).

% rel_tos1(ST,X,F,Xold,Old,New),
% rel_tos1: Stack x X x F x Old x New
% Xold = matching SO
% X = updated X
% Old = actual stack item (may include wrapping once(_) etc.)
% New = Old with updated occ

rel_tos1([TOP|L],X,F,Xold,Old,New) :-
    (functor(TOP,inactive,1) ->	% bypass inactive(_)
	 !,			% red
	 rel_tos1(L,X,F,Xold,Old,New)
     ;  (boundaryMark(TOP) ->	% look through boundary if global access
	     !, 		% red
	     stack_access(global),
	     rel_tos1(L,X,F,Xold,Old,New))).
rel_tos1([once(X)|_],Xnew,F,X,once(X),once(Xnew)) :-
    checkfeature(X,F),
    !,				% red
    new_occ(X,Xnew).
rel_tos1([X|_],Xnew,F,X,X,Xnew) :-
    checkfeature(X,F),
    !,				% red
    new_occ(X,Xnew).
rel_tos1([SK|L],X,F,Xold,Old,New) :-
    printStackSkip(SK,F),
    rel_tos1(L,X,F,Xold,Old,New).

% relativized TOS with ID
% rel_tos(STK,X,ID,F,Old,New) finds first stack element with F and ID
% ID is matched for nom(ID) Case assignment (for locality)
% respects inactive(_) wrapper unless parameter c_stack_search(inactive) is set.
% stack_access parameter:
%  local: can't look past b
%  global: skips b markers
% spellout: new occ, old occ
% old copy on stack is removed
% new copy has fresh occ, same fresh occ copy is returned as X
% also: old element gets occ valued (so we know it's not to be spelled out)
rel_tos(ST,X,ID,F,STp) :-
    rel_tos1(ST,X,ID,F,Xold,Old,New),
    updateStack(ST,Xold,X,Old,New,STp).

rel_tos1([TOP|L],X,ID,F,Xold,Old,New) :-
    ((c_stack_search(normal),
      functor(TOP,inactive,1)) ->
	 !,			% red
	 rel_tos1(L,X,ID,F,Xold,Old,New)
     ;  (boundaryMark(TOP) ->
	     !, 		% red
	     stack_access(global),
	     rel_tos1(L,X,ID,Xold,Old,New))).
rel_tos1([IX|_],Xnew,ID,F,X,IX,inactive(Xnew)) :-
    functor(IX,inactive,1), arg(1,IX,X),
    checkfeature(X,ID,F),
    !,				% red
    new_occ(X,Xnew).
rel_tos1([once(X)|_],Xnew,ID,F,X,once(X),once(Xnew)) :-
    checkfeature(X,ID,F),
    !,				% red
    new_occ(X,Xnew).
rel_tos1([X|_],Xnew,ID,F,X,X,Xnew) :-
    checkfeature(X,ID,F),
    !,				% red
    new_occ(X,Xnew).
rel_tos1([X|_],Xnew,ID,F,X,X,Xnew) :-
    complement(X,Y), checkfeature(Y,ID,F),
    !,				% red
    new_occ(X,Xnew).
rel_tos1([SK|L],X,ID,F,Xold,Old,New) :-
    printStackSkip(SK,F),
    rel_tos1(L,X,ID,F,Xold,Old,New).

% no copy version of rel_tos/4
rel_tos_ncp([TOP|L],X,ID,F) :-
    ((c_stack_search(normal),
      functor(TOP,inactive,1)) ->
	 !,			% red
	 rel_tos_ncp(L,X,ID,F)
     ;  (boundaryMark(TOP) ->
	     !, 		% red
	     stack_access(global),
	     rel_tos_ncp(L,X,ID,F))).
rel_tos_ncp([IX|_],X,ID,F) :-
    functor(IX,inactive,1), arg(1,IX,X),
    checkfeature(X,ID,F),
    !.				% red
rel_tos_ncp([once(X)|_],X,ID,F) :-
    checkfeature(X,ID,F),
    !.				% red
rel_tos_ncp([X|_],X,ID,F) :-
    checkfeature(X,ID,F),
    !.				% red
%%% NEW: also allow recursively: of who_rel person's dog's mother
rel_tos_ncp([X|_],X,ID,F) :-
	complement(X,Y),
	(checkfeature(Y,ID,F) ;
	 is_phase(Y),
	 recursive_edge(Y,Z),
	 checkfeature(Z,ID,F)),
    !.				% red
%%% NEW: allow feature to be in specifier of phase
%%% also allow recursively: who_rel person's dog's mother
rel_tos_ncp([X|_],X,ID,F) :-
	is_phase(X),
	recursive_edge(X,Y),
	checkfeature(Y,ID,F),
	!.			% red
rel_tos_ncp([SK|L],X,ID,F) :-
    printStackSkip(SK,F),
    rel_tos_ncp(L,X,ID,F).

recursive_edge(X,Y) :- in_edge(X,Y).
recursive_edge(X,Y) :- in_edge(X,Z), recursive_edge(Z,Y).

% looks for subconstituent of X in the stack
% assumes stack element is marked with feature substream(_)
% if it came from a substream
% Possible Behaviors: subphrase_check parameter:
%   subsume:	subconstituent is subsumed
%   on_fail: 	fail if subconstituent is found
%   subsume_ss: only a subconstituent from a substream is subsumed
%   on_fail_ss: fail if subconstituent from a substream is found
%   off: no check
%
% subsumption works for subject extraction cases (subsume_ss)
% but for relative clauses, it's needed: the man who loves mary
% who-man should not result in 'man' being deleted in subsume_ss mode
% HACK: skip heads...

% Another HACK
% subsumption of "who_rel man" for "friends of who_rel man" cannot work
% otherwise, who_rel man can never be extracted if deleted from the stack

checkSubconstituent(X,S,Sp) :-
    (subphrase_check(off) ->
	 fail
     ;  checkSC(S,X,Sp)).

checkSC([X|L],Y,S) :-
    (boundaryMark(X) ->
	 S = [X|Sp],
	 checkSC(L,Y,Sp)
     ;  (head(X) ->		% hack
	     S = [X|Sp],
	     checkSC(L,Y,Sp)
	 ; ((functor(X,F,1), (F = once ; F = inactive), arg(1,X,X2)) ->
		subC(X2,Y,SS),
		subphrase_check_behavior(SS,X2,Y,L,S)
	    ;  subC(X,Y,SS),
	       subphrase_check_behavior(SS,X,Y,L,S)))).

% X is a proper subconstituent of Y
% 3rd argument return value: [] (same stream), ss(_) (X is from a substream)
proper_subC(X,Y,A) :-
    subC(X,Y,A), \+ X = Y.
		 
% X is a subconstituent of Y
% 3rd argument return value: [] (same stream), ss(_) (X is from a substream)
subC(X,Y,[]) :-
\+ \+ (X = so(LB,U,I,F,Cs), Y = so(LB,U,I,F,Cs)),
!.				% red
subC(X,Y,ss(V)) :-		% V will not be instantiated!
    \+ \+ (
	X = so(LB,U,I,FX,Cs), Y = so(LB,U,I,FY,Cs),
	pick(substream(V),FX,FY)),
    !.		% red
subC(X,Y,SS) :- c(Y,Z), subC(X,Z,SS). 

subphrase_check_behavior(_SS,X,_,_,_) :-
	iF(X,f(rel)),		% need to be able to raise rel
	!,			% needed
	fail.
subphrase_check_behavior(_SS,X,_,_,_) :-
	label(X,type(p)),
	complement(X,Y),
	iF(Y,f(rel)),		% pied-piped version of above
	!,
	fail.
subphrase_check_behavior(_SS,X,Y,L,L) :- % remove X
    subphrase_check(subsume),
    !,
    printSubCSubsumed(X,Y).
subphrase_check_behavior(_,X,Y,_,_) :-
    subphrase_check(on_fail),
    !,
    printFailSubC(X,Y),
    fail. 
subphrase_check_behavior(ss(_),X,Y,L,L) :-  % remove X
    subphrase_check(subsume_ss),
    !,
    printSubCSSSubsumed(X,Y).
subphrase_check_behavior(ss(_),X,Y,_,_) :-  
    subphrase_check(on_fail_ss),
    !,
    printFailSubCSS(X,Y),
    fail.  

stack_depth(S,D) :- length(S,D).

%%% PRINTING

% uses portray/1
printState(Str,Stk,SO,Step) :-
    inc_stats_value(steps,V),
    stack_depth(Stk,N),
    graph(N),
    (N>1 -> M is N-1, inc_stats(depth,M) ; true),
    step_lastvalue(LV),
    (LV > Step -> true ; LV = Step),
    printState1(Str,Stk,SO,LV),
    (true; backtrackstep(V), fail).

printState1(Str,Stk,SO,_Step) :-
    type(text),
    !,				% red
    format(string(Out1),'Stream:~p',[Str]),
    output(Out1),
\+ \+ ((iprint(off) -> iPrintL(Stk,Stkp) ; Stkp = Stk),
       format(string(Out2),'Stack: ~p',[Stkp]), output(Out2)),
    \+ \+ (format(string(Out3),'SO:    ~p',[SO]), output(Out3)),
    output('').
printState1(Str,Stk,SO,Step) :-
%    copy_term(Stk,Stk1), copy_term(SO,SO1),
    format(string(Out1),'<table class=ps id=step~w><tr><td class=stat>Stream:</td><td class=stat>~p</td></tr>',
	   [Step,Str]),
    \+ \+ (stackVPrint(Stk,Out2),
	   format(string(Out3),'<tr><td class=stat>SO:</td><td class=so>~p<br></td></tr></table>',[SO]),
	   string_concat(Out1,Out2,Out12),
	   string_concat(Out12,Out3,Out123),
	   output(Out123)),
    draw_tree(SO,Step).

% sensitive display parameter iprint inactive(_)
stackVPrint(L,Out) :-
    vprint(off),		% display parameter
    !,				% red
    (iprint(off) ->		% display parameter
	 iPrintL(L,Lp)
     ;   Lp = L),
    format(string(Out),
	   '<tr><td class=stat>Stack:</td><td class=stat>~p</td></tr>',[Lp]).
stackVPrint(L,Out) :-
    format(string(Out1),'<tr><td class=stat>Stack (⥥):</td><td class=stat>',[]),
    format(string(Out2),'</td></tr>',[]),
    (iprint(off) ->		% display parameter
	 loopPrintOLi(L,OutL,V)
     ;   loopPrintOL(L,OutL,V)),
    V = [Out2],
    atomics_to_string([Out1|OutL],Out).

%%% NEW: stack element Labeling check
loopPrintOL([],L,L).
loopPrintOL([X|L],[Out|L1],L2) :-
	(labelingCheck(X,Rep) ->
	 (Rep == '' -> Out = '' ; format(string(Out),'~p<br>',[Rep]))
	;
	 format(string(Out),'~p<br>',[X])),
	loopPrintOL(L,L1,L2).

labelingCheck(X,Out) :-
	functor(X,c,2),
	arg(1,X,X3),
	arg(2,X,Goal),
	(\+ \+ Goal -> Out = '' ; atom_concat(X3,'?',Out)).

% inactive(_) wrapper deletion version of loopPrintOL/3
%%% NEW: stack element Labeling check
loopPrintOLi([],L,L).
loopPrintOLi([X|L],[Out|L1],L2) :-
    (functor(X,inactive,1) ->
     arg(1,X,X2),
     format(string(Out),'~p<br>',[X2])
    ; (labelingCheck(X,Rep) ->
       (Rep == '' -> Out = '' ; format(string(Out),'~p<br>',[Rep]))
      ;
       format(string(Out),'~p<br>',[X]))),
    loopPrintOLi(L,L1,L2).

% delete inactive(_) wrappers
iPrintL([],[]).
iPrintL([X|L],[Y|Lp]) :-
	labelingCheck(X,Y),
	!,			% red
	iPrintL(L,Lp).
iPrintL([X|L],[Y|Lp]) :-
	rmInactive(X,Y),
	iPrintL(L,Lp).

rmInactive(X,Y) :-
    functor(X,inactive,1) -> arg(1,X,Y) ; Y = X.

printRetry(Step) :-
    format(string(Out),'Retry step ~p: instead of pair merge...',[Step]),
    output(Out).

printLabelingFailure(C) :-
    format(string(Out),'XP-YP labeling failure at node ~p?',[C]),
    output(Out).

printInitial(X) :-
    format(string(Out),'Head of stream ~p is the initial SO',[X]),
    output(Out).

printInternalMerge :-
    format(string(Out),'Internal merge selected',[]),
    output(Out).

printMerge(X,Y,H,Txt) :-
    inc_stats(merge),
    format(string(Out),'Merge ~p and ~p',[abbrv(X),abbrv(Y)]),
    output(Out),
    format(string(Out2),'Label from ~p (~w)',[abbrv(H),Txt]),
    output(Out2).

printMergeCond(X,Y,H,Txt) :-
    inc_stats(merge),
    format(string(Out),'Merge ~p and ~p',[abbrv(X),abbrv(Y)]),
    output(Out),
    format(string(Out2),'Conditional label from ~p (~w)',[abbrv(H),Txt]),
    output(Out2).

printPairMerge(X,Y) :-
    inc_stats(merge),
    format(string(Out),'Pair-merge ~p and ~p',[X,Y]),
    output(Out).

printPushNoPiedPipe(X) :-
    format(string(Out),'Pair-merge adjunct ~p does not pied-pipe',[abbrv(X)]),
    output(Out).

printPush(X) :-
    inc_stats(push),
    format(string(Out),'Push ~p (unvalued uF) onto stack',[abbrv(X)]),
    output(Out).

printFront(X) :-
    inc_stats(push),
    inc_stats(nottos),
    format(string(Out),
	   'Move ~p (unvalued uF) to the front of the stack',
	   [abbrv(X)]),
    output(Out).

printPop(X) :-
    inc_stats(pop),
    format(string(Out),'Pop ~p (no unvalued/un-unified uF)',[abbrv(X)]),
    output(Out).

printProbe(P,F) :- 
    (var(F) ->
	 format(string(Out),'~p probes with edge feature',[abbrv(P)])
     ;  format(string(Out),'~p probes with ~w-edge feature',[abbrv(P),F])),
    output(Out).

printProbe(P) :-
    ((uF(P,f(phi,_,_)), \+ phi_complete(P)) ->
	 format(string(Out),'Defective ~p probes',[abbrv(P)])
     ;  format(string(Out),'~p probes',[abbrv(P)])),
    output(Out).

printNospell(X,Y) :-
    format(string(Out),'Fail: ~p ~p is a lexical gap',[X,Y]),
    output_error(Out).

printDropped(L) :-
    (L = [_|_] ->
	 length(L,N),
	 inc_stats(pop,N),
	 format(string(Out),
		'Drop ~p from the stack (2 boundaries away)',[L]),
	 output(Out)
     ;   true).

printValueCaseFail(G,V) :-
    format(string(Out),'Fail: goal ~p has ~w Case',[G,V]),
    output_error(Out).

printValue(X,G,F) :-
    format(string(Out),'~p values u~w on ~p',[X,F,G]),
    output(Out).

printEdgePreference(X) :-
    head(X,H),
    format(string(Out),'Preference: ~p edge merge from the stack (not stream)',[H]),
    output(Out).

printSSOInsert(X) :-
    format(string(Out),'End substream, SO ~p re-inserted into main stream',[X]),
    output(Out).

printSSInsert(S) :-
    format(string(Out),'Substream stack ~p saved for possible stack merge',[S]),
    output(Out).

printMergeDelayedStack(Stack) :-
	format(string(Out),'Saved substream stack ~p merged with stack',[Stack]),
	output(Out).

printStackLE :-
    inc_stats(b),
    format(string(Out),'New boundary (b) marker stacked',[]),
    output(Out).

printStackSkip(X,F) :-
    inc_stats(nottos),
    format(string(Out),'Relativized TOS search for ~w skipping ~p',[F,abbrv(X)]),
    output(Out).

printStackSearchInactive(X) :-
    (iprint(on) ->		% display parameter
	 format(string(Out),'Inactive stack element ~p found',[abbrv(X)])
     ;   rmInactive(X,Xp),
         format(string(Out),'Unlicensed stack element ~p found',[abbrv(Xp)])),
    output(Out).

printStackSearchUF(X) :-
    format(string(Out),'Stack element ~p with unvalued uF found',
	   [abbrv(X)]),
    output(Out).
    
printStackFronting(X) :-
    format(string(Out),'First goal ~p stack fronted',[abbrv(X)]),
    output(Out).

printUndoSStackMerge(L) :-
	format(string(Out),
	       'Pair-merge: substream stack merge of ~p canceled',[L]),
	output(Out).

printFailFormat(FM,L) :-
    format(string(Out),FM,L),
    output_error(Out).

printFail(F) :-
    format(string(Out),F,[]),
    output_error(Out).

printSubCSubsumed(X,Y) :-
    format(string(Out),
	   'Subconstituent ~p removed from stack, ~p subsumes it',
	   [abbrv(X),abbrv(Y)]),
    output_error(Out).

printSubCSSSubsumed(X,Y) :-
    format(string(Out),
	   'Subconstituent ~p from a substream removed from stack, ~p subsumes it',
	   [abbrv(X),abbrv(Y)]),
    output_error(Out).

printFailSubC(X,Y) :-
    format(string(Out),
	   'Fail: ~p has subconstituent ~p already on the stack',
	   [abbrv(Y),abbrv(X)]),
    output_error(Out).

printFailSubCSS(X,Y) :-
    format(string(Out),
	   'Fail: ~p has subconstituent ~p from a substream already on the stack',
	   [abbrv(Y),abbrv(X)]),
    output_error(Out).

printFailStack(X) :-
    format(string(Out),'Fail: ~p has uninterpretable feature(s) but there are too many intervening stack Local Extent boundaries',[X]),
    output_error(Out).

printAttach(X,Dir) :-
    format(string(Out),'~p has feature attach(~w)',[X,Dir]),
    output(Out).

printProbegoal(X,Y,Txt) :-
    inc_stats(agree),
    format(string(Out),'~p ~w ~p',[abbrv(X),Txt,abbrv(Y)]),
    output(Out).

printProbegoal(X,Y,Txt1,Z,Txt2) :-
    inc_stats(agree),
    format(string(Out),'~p ~w ~w ~w ~p',[abbrv(X),Txt1,Z,Txt2,abbrv(Y)]),
    output(Out).

printProbegoalEconomy(X,Y,Txt1,Z,Txt2) :-
    inc_stats(agree),
    format(string(Out),'~p ~w ~w ~w ~p (economy)',
	   [abbrv(X),Txt1,Z,Txt2,abbrv(Y)]),
    output(Out).

printProbegoalUnify(P,G,Unified) :-
    inc_stats(unify),
    format(string(Out),'Unified ~w feature on ~p and ~p',[Unified,abbrv(P),abbrv(G)]),
    output(Out).

printInherit(X,Is,_Fs) :-
    prettify_list(Is,Isp),
    format(string(Out),
	   'Inherit interpretable feature(s) ~w from ~p',[Isp,abbrv(X)]),
    output(Out).

printRelease(X) :-
    inc_stats(release),
    format(string(Out),'Restack ~p onto stack',[abbrv(X)]),
    output(Out).
    
printPhaseCompleted(X) :-
    format(string(Out),'Local Extent boundary at ~p',[abbrv(X)]),
    output(Out).

printThetaMarked(X) :-
    format(string(Out),'Theta-mark ~p',[abbrv(X)]),
    output(Out).

printCompoundingValuation(X) :-
	format(string(Out),
	       'Noun-noun compounding, valued uD on ~p',[abbrv(X)]),
	output(Out).

printPMInherentCase(XP) :-
	format(string(Out),'Inherent Case valued on ~p',[abbrv(XP)]),
	output(Out).

printPMNoTheta(XP) :-
	format(string(Out),'Pair Merged ~p has no theta role',[abbrv(XP)]),
	output(Out).

printPMTheta(XP) :-
	format(string(Out),'Pair Merged ~p has inherent theta role',[abbrv(XP)]),
	output(Out).

printNoAction(S,SO) :-
    format(string(Out),
	   'No possible action for stream ~p and ~p',[S,abbrv(SO)]),
    output(Out).

printFormSet(XP,SO) :-
	(SO = set(_) ->
	 format(string(Out),'FormSet: adding ~p to set ~p',[abbrv(XP),abbrv(SO)])
	;
	 format(string(Out),'FormSet: set members ~p and ~p',[abbrv(XP),abbrv(SO)])),
	output(Out).

% statistics

init_stats :-
	nb_setval(unify,0),
	nb_setval(merge,0),
	nb_setval(agree,0),
	nb_setval(push,0),
	nb_setval(pop,0),
	nb_setval(nottos,0),
	nb_setval(release,0),
	nb_setval(b,0),
	nb_setval(depth,0),
	nb_setval(steps,-1),
	nb_setval(graph,[]).

inc_stats(X) :-	nb_getval(X,N), M is N+1, nb_setval(X,M).
inc_stats_value(X,M) :-	nb_getval(X,N), M is N+1, nb_setval(X,M).
inc_stats(X,K) :- nb_getval(X,N), M is N+K, nb_setval(X,M).

print_stats :-
	nb_getval(unify,U),
	nb_getval(merge,M),
	nb_getval(agree,A),
	nb_getval(push,P1),
	nb_getval(pop,P2),
	nb_getval(nottos,P3),
	nb_getval(release,P4),
	nb_getval(b,P5),
	nb_getval(depth,D),	
	format(string(Out1),
'Merge: ~w, Agree: ~w, Unify: ~w, Push: ~w, Boundaries: ~w, Inject: ~w',
	       [M,A,U,P1,P5,P4]),
	output(Out1),
	format(string(Out2),
	       'Non-TOS access: ~w, Pop: ~w, Depth cost: ~w',
	       [P3,P2,D]),
	output(Out2),
	print_graph.

graph(N) :-
	nb_getval(steps,S),
	nb_getval(graph,L),
	nb_setval(graph,[S,N|L]).

% derivation stack depth chart
print_graph :-
    type(text),
    !,				% red
    nb_getval(graph,L),
    with_output_to(string(Out),
		   (format('Step:       '),
		    print_graph1(L))),
    output(Out),
    with_output_to(string(Out2),
		   (format('Stack depth:'),
		    print_graph2(L))),
    output(Out2).
print_graph :-
    nb_getval(graph,L),
    with_output_to(string(Out),
		   (format('<table><tr><td>Step:</td>'),
		    print_graph1_html(L),
		    format('</tr><tr><td>Stack depth:</td>'),
		    print_graph2_html(L))),
    output(Out).

print_graph1([]).
print_graph1([S,_N|L]) :- print_graph1(L), format('~|~t~d~3+',[S]).

print_graph1_html([]).
print_graph1_html([S,_N|L]) :- print_graph1_html(L), format('<td class=stat>~d</td>',[S]).

print_graph2([]).
print_graph2([_S,N|L]) :- print_graph2(L), format('~|~t~d~3+',[N]).

print_graph2_html([]).
print_graph2_html([_S,N|L]) :- print_graph2_html(L), format('<td class=stat>~d</td>',[N]).

resetStep :- flag(step,_,1).
step(X) :- 
    flag(step,X,X), 
    format(string(Out),'Step ~w. ',[X]), 
    output(Out,h3),
    Y is X+1, flag(step,_,Y).

step_lastvalue(Y) :- flag(step,X,X), Y is X-1.

% parse canvas count
resetC :- flag(parse,_,0).
newC(X) :- flag(parse,X,X), Y is X+1, flag(parse,_,Y).
parseC(X) :- flag(parse,X,X).

% derivation counting
% not used for single runs
resetD :- flag(deriv,_,0).
bumpD :-
    flag(deriv,X,X), Y is X+1, flag(deriv,_,Y),
    set_deriv(Y).
derivD(X) :- flag(deriv,X,X).

%% machine parameterization
print_parameters :-
    type(text),
    !,				% red
    print_parameters_text.
print_parameters :-
    findall((X,Y),parameterization(X,Y),L),
    with_output_to(string(Out),
		   (format('<table><thead class=fixedHeader><tr><th>Parameter</th><th>Setting</th><th>Description</th></tr></thead><tbody class=scrollContent>'),
		    print_parametersL(L),
		    format('</tbody></table>'))),
    output_settings(Out).

print_parameters_text :-
    format('~nParameter~17|Setting~11+Description~n',[]),
    parameterization(P,Txt),
    parameter_name(P,N),
    parameter_value(P,V),
    format('~w~17|~w~11+~w~n',[N,V,Txt]),
    fail.
print_parameters_text.

parameter_name(A,A) :-
    atom(A),
    !.				% red
parameter_name(T,F) :-
    functor(T,F,_).
    
parameter_value(A,V) :-
    atom(A),
    !,				% red
    (call(A) -> V = on ; V = off).
parameter_value(T,V) :-
    call(T),
    !,				% red
    arg(1,T,V).
parameter_value(_,unknown).

% code for parameter P's current value
parameter_code(P,Code) :-
    parameterization(P,_),
    !,				% red
    parameter_value(P,V),
    parameter_code2(P,V,Code).
parameter_code(F,Code) :-
    parameterization(P,_),
    functor(P,F,1),
    !,				% green, red (multiple defs)
    parameter_value(P,V),
    parameter_code2(P,V,Code).

% code for parameter P's next value
parameter_code_next(P,Code) :-
    parameterization(P,_),
    !,				% red
    next_parameter_value(P,V2),
    parameter_code2(P,V2,Code).
parameter_code_next(F,Code) :-
    parameterization(P,_),
    functor(P,F,1),
    !,				% green, red (multiple defs)
    next_parameter_value(P,V2),
    parameter_code2(P,V2,Code).

parameter_code2(A,on,A) :- atom(A), !. % red
parameter_code2(A,off,(A :- fail)) :- atom(A), !. % red
parameter_code2(P,V,P2) :- functor(P,F,1), functor(P2,F,1), arg(1,P2,V).

next_parameter_value(P,V2) :-
    parameter_value(P,V),
    (atom(P) ->
	 (V == on -> V2 = off ; V2 = on)
     ;  parameter_values(P,L),
	nextInRing(L,V,V2)).

nextInRing(L,V,V2) :-
    append(Left,[V|Right],L),
    (Right == [] ->
	 (Left \== [] ->
	      Left = [V2|_]
	  ;  V2 = V)
     ;  Right = [V2|_]),
    !.				% green

print_parametersL([]).
print_parametersL([(P,Txt)|L]) :-
    parameter_name(P,N),
    parameter_value(P,V),
    format('<tr class=nrow><td>~w</td><td class=is onClick=toggle(this)>~w</td><td>~w</td></tr>',[N,V,Txt]),
    !, 				% green
    print_parametersL(L).

toggle_parameter(N) :-
    parameter_code(N,C),
    parameter_code_next(N,C2),
    retract(C),
    assert(C2),
    print_parameters.

% debugging
% uses global fail_chain to prevent printing of more error messages on backtrack

succeed(G,S,F) :-
    (call(G) *->
	 printSuccess(S)
     ;  (F = [FM,L] ->
	     printFailFormat(FM,L)
	 ;   printFail(F)),
	fail).
%	 nb_setval(fail_chain,1),

% quiet fail version
succeed(G,S) :-
    (call(G) *->
	 printSuccess(S)
     ;  fail).

%% Hacks

released(X) :- f(X,release(V)), nonvar(V).

% works (unintentional?) even if X is a variable, not a SO
mark_releasedNF(X) :-
    search(so),
    !,				% red
    mark_release(X).
mark_releasedNF(_).

mark_released(X) :- f(X,release(V)), var(V), !, V = r.

% adds once(_) wrapper to Y if X has feature release(_)
add_wrapper(X,Y,once(Y)) :-
    search(stk),
    f(X,release(_)),
    !.				% red
add_wrapper(_,Y,Y).

% list utilities

pick(X,[X|L],L).
pick(X,[Y|L],[Y|Lp]) :- pick(X,L,Lp).
    
butlast([_],[]).
butlast([X|L],[X|Lp]) :- butlast(L,Lp).

extendOL(V) :- var(V) -> V=[_|_] ; ov(V,OV), OV =[_|_].

ov(V,OV) :- var(V) -> OV = V ; V = [_|V2], ov(V2,OV).

lengthOL(V,N) :- var(V) -> N = 0 ; V = [_|V2], lengthOL(V2,M), N is M+1.
